import time

import cherrypy
from cherrypy import Tool, request, tools

from WMCore.REST.Auth import RESTAuth


def set_request_time():
    """Utility to time stamp the start of HTTP request handling."""
    request.start_time = time.time()


def set_proxy_base(base=None):
    """Utility to correctly handle requests behind a proxy."""
    scheme = request.headers.get('X-Forwarded-Proto', request.base[:request.base.find("://")])
    base = request.headers.get('X-Forwarded-Host', base)
    if not base:
        port = cherrypy.request.local.port
        if port == 80:
            base = 'localhost'
        else:
            base = 'localhost:%s' % port

    base = base.split(',')[0].strip()
    if base.find("://") == -1:
        base = scheme + "://" + base
    request.base = base

    xff = request.headers.get('X-Forwarded-For')
    if xff:
        xff = xff.split(',')[0].strip()
        request.remote.ip = xff


tools.cms_auth = RESTAuth()
tools.time = Tool('on_start_resource', set_request_time)
tools.proxy = Tool('before_request_body', set_proxy_base, priority=30)



"""
_GetDeletedBlocksByWorkflow_

MySQL implementation of Workflows.GetDeletedBlocksByWorkflow

Retrieves a list of workflows with lists of deleted and NOT deleted blocks per workflow,

NOTE: This DAO is not used in the production code but is to be used only for debugging purposes
"""


from WMCore.Database.DBFormatter import DBFormatter


class GetDeletedBlocksByWorkflow(DBFormatter):
    """
    Retrieves a list of all workflows and the relative deleted blocks lists
    """
    sql = """SELECT
                 dbsbuffer_block.blockname,
                 dbsbuffer_block.deleted,
                 wmbs_workflow.name
             FROM dbsbuffer_block
             INNER JOIN dbsbuffer_file ON
                 dbsbuffer_file.block_id = dbsbuffer_block.id
             INNER JOIN dbsbuffer_workflow ON
                 dbsbuffer_workflow.id = dbsbuffer_file.workflow
             INNER JOIN wmbs_workflow ON
                 wmbs_workflow.name = dbsbuffer_workflow.name
             GROUP BY
                 dbsbuffer_block.blockname,
                 dbsbuffer_block.deleted,
                 wmbs_workflow.name
          """

    def format(self, result):
        """
        _format_

        Format the query results into the proper dictionary expected at the upper layer Python code.
        The input should be a list of database objects, each one representing a single line returned
        from the database with key names matching the column names from the sql query.

        The intermediate (not aggregated) result representing the primary database output in python should be
        a list of dictionaries one record per line returned from the database with key names mapped to the
        python code variable naming conventions.

        e.g.
        [{'blockname': '/a/b/c#123-qwe',
          'deleted': 0,
          'name': 'WorkflowName'},
        {'blockname': '/a/b/c#456-rty',
          'deleted': 1,
          'name': 'WorkflowName'},
         {'blockname': '/a/b/d#123-asd',
          'deleted': 0,
          'name': 'WorkflowName'}
        ...
        ]

        NOTE:
         * The number of records per workflow and block returned (i.e. number of records per group in the GROUP BY statement)
           from the query is not related to either the number of blocks nor to the number of workflows, but rather to the
           combination of number of files in the block and some other factor which increases the granularity (it seems to be
           the number of records in dbsbuffer_workflow table per file aggregated by workflow), and NO `DISTINCT` requirement
           in the SELECT statement is needed because we already have them properly grouped.
         * Once deleted we should NOT expect duplicate records with two different values of the deleted
           flag to be returned for a single block but we should still create the list of deleted and
           NotDeleted blocks as sets and eventually create their proper intersection for double check.

        This list needs to be further aggregated by name to produce an aggregated structure per workflow like:

        [{'name': 'WorkflowName'
          'blocksNotDeleted': ['/a/b/c#123-qwe',
                               /a/b/c#456-rty']
          'blocksDeleted': ['/a/b/d#123-asd']
         },
        ...
        ]

        :param result: The result as returned by the mysql query execution.
        :return:       List of dictionaries
        """

        # First reformat the output in a list of dictionaries per DB record
        dictResults = DBFormatter.formatDict(self, result)

        # Now aggregate all blocks per workflow:
        results = {}
        for record in dictResults:
            wfName = record['name']
            results.setdefault(wfName, {'name': wfName, 'blocksDeleted': [], 'blocksNotDeleted': []})
            if record['deleted']:
                results[wfName]['blocksDeleted'].append(record['blockname'])
            else:
                results[wfName]['blocksNotDeleted'].append(record['blockname'])
        return results.values()

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        Executing the current sql query.
        :param conn:        A current database connection to be used if existing
        :param transaction: A current database transaction to be used if existing
        :return:            A list of dictionaries one record for each database line returned
        """
        results = self.dbi.processData(self.sql, conn=conn,
                                       transaction=transaction)

        return self.format(results)



"""
File       : MSPileup.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSPileup provides logic behind the pileup WMCore module.
"""

# system modules
from threading import current_thread

# WMCore modules
from WMCore.MicroService.MSCore.MSAuth import MSAuth
from WMCore.MicroService.MSCore.MSCore import MSCore
from WMCore.MicroService.DataStructs.DefaultStructs import PILEUP_REPORT
from WMCore.MicroService.MSPileup.MSPileupData import MSPileupData


class MSPileup(MSCore):
    """
    MSPileup provides whole logic behind the pileup WMCore module.
    """

    def __init__(self, msConfig, **kwargs):
        super().__init__(msConfig)
        self.dataMgr = MSPileupData(msConfig)
        self.authMgr = MSAuth(msConfig)
        # Get the RSE expression for Disk RSEs from the configuration
        self.diskRSEExpr = msConfig.get("rucioDiskExpression", "")

    @staticmethod
    def status():
        """
        Provide MSPileup status API. We should extend it to check DB connection, etc.

        :return: status dictionary
        """
        summary = dict(PILEUP_REPORT)
        summary.update({'thread_id': current_thread().name})
        return summary

    def userDN(self):
        """
        Return user DN from authentication manager
        :return: string
        """
        user = self.authMgr.userInfo()
        dn = user.get('dn', 'Unknown')
        return dn

    def getPileup(self, **kwargs):
        """
        MSPileup get API fetches the data from underlying database layer

        :param **kwargs: provide key=value (or spec) input
        :return: results of MSPileup data layer (list of dicts)
        """
        spec = {}  # (get all full docs) retrieve a list with the full documentation of all the pileups
        if 'query' in kwargs:
            # use specific spec (JSON query)
            spec = kwargs['query']
        elif 'pileupName' in kwargs:
            # get docs for given pileupName
            spec = {'pileupName': kwargs['pileupName']}
        elif 'campaign' in kwargs:
            # get docs for given campaign
            spec = {'campaign': kwargs['campaign']}
        else:
            for key, val in kwargs.items():
                spec[key] = val

        # check if filters are present and use it as projection fields
        projection = {}
        filters = kwargs.get('filters', [])
        # convert a single filters to a list
        if isinstance(filters, str):
            filters = [filters]
        for key in kwargs.get('filters', []):
            # filter out empty strings
            if key:
                projection[key] = 1
        results = self.dataMgr.getPileup(spec, projection)

        return results

    def queryDatabase(self, query, projection=None):
        """
        MSPileup query database API querying the data in underlying data layer.

        :param query: provide query JSON spec to MSPileup data layer
        :return: results of MSPileup data layer (list of dicts)
        """
        spec = {}  # (get all full docs) retrieve a list with the full documentation of all the pileups
        if 'query' in query:
            # use specific spec (JSON query)
            spec = query['query']
        # check if filters are present and use it as projection fields
        projection = {}
        for key in query.get('filters', []):
            projection[key] = 1
        return self.dataMgr.getPileup(spec, projection)

    def createPileup(self, pdict):
        """
        MSPileup create pileup API to create appropriate pileup document
        in underlying database.

        :param pdict: input MSPileup data dictionary
        :return: results of MSPileup data layer (list of dicts)
        """
        self.authMgr.authorizeApiAccess('ms-pileup', 'create')
        rseNames = self.rucio.evaluateRSEExpression(self.diskRSEExpr, useCache=True)
        return self.dataMgr.createPileup(pdict, rseNames, userDN=self.userDN())

    def updatePileup(self, pdict):
        """
        MSPileup update API to update corresponding pileup document in data layer.

        :param pdict: input MSPileup data dictionary, e.g. partial pileup spec
        {"containerFraction": float, "pileupName": string}.
        :return: results of MSPileup data layer (list of dicts)
        """
        self.authMgr.authorizeApiAccess('ms-pileup', 'update')
        rseNames = self.rucio.evaluateRSEExpression(self.diskRSEExpr, useCache=True)
        return self.dataMgr.updatePileup(pdict, rseNames, validate=True, userDN=self.userDN())

    def deletePileup(self, spec):
        """
        MSPileup delete API to delete corresponding pileup document in data layer.

        :param pdict: input MSPileup data dictionary
        :return: results of MSPileup data layer (list of dicts)
        """
        self.authMgr.authorizeApiAccess('ms-pileup', 'delete')
        return self.dataMgr.deletePileup(spec)



"""
File       : MSManager.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
             Alan Malta <alan dot malta AT cern dot ch >
Description: MSManager class provides full functionality of the MSManager service.
It provides a interface to reqmgr2ms service and should be
used in service config.py as following

.. doctest::

    # REST interface
    data = views.section_('data')
    data.object = 'WMCore.MicroService.Service.RestApiHub.RestApiHub'
    data.manager = 'WMCore.MicroService.MSManager.MSManager'
    data.reqmgr2Url = "%s/reqmgr2" % BASE_URL
    data.limitRequestsPerCycle = 500
    data.enableStatusTransition = False
    data.enableDataTransfer = False
    data.verbose = True
    data.interval = 60
    data.rucioAccount = RUCIO_ACCT
    data.dbsUrl = "%s/dbs/%s/global/DBSReader" % (BASE_URL, DBS_INS)
"""
# futures
from __future__ import division, print_function

# system modules
from builtins import object
from time import sleep
from datetime import datetime

# WMCore modules
from WMCore.MicroService.Tools.Common import getMSLogger
from WMCore.MicroService.MSCore.TaskManager import start_new_thread
from Utils.Utilities import strToBool

def daemon(func, reqStatus, interval, logger):
    "Daemon to perform given function action for all request in our store"
    while True:
        try:
            func(reqStatus)
        except Exception as exc:
            logger.exception("MS daemon error: %s", str(exc))
        sleep(interval)


def daemonOpt(func, interval, logger, *args, **kwargs):
    "Daemon to perform given function action for all request in our store"
    while True:
        try:
            func(*args, **kwargs)
        except Exception as exc:
            logger.exception("MS daemon error: %s", str(exc))
        sleep(interval)


class MSManager(object):
    """
    Entry point for the MicroServices.
    This class manages both transferor and monitoring services.
    """

    def __init__(self, config=None, logger=None):
        """
        Initialize MSManager class with given configuration,
        logger, ReqMgr2/ReqMgrAux/PhEDEx/Rucio objects,
        and start transferor and monitoring threads.
        :param config: reqmgr2ms service configuration
        :param logger:
        """
        self.config = config
        self.logger = getMSLogger(getattr(config, 'verbose', False), logger)
        self._parseConfig(config)
        self.logger.info("Configuration including default values:\n%s", self.msConfig)
        self.statusTrans = {}
        self.statusMon = {}
        self.statusOutput = {}
        self.statusRuleCleaner = {}
        self.statusUnmerged = {}
        self.statusPileup = {}

        # initialize pileup module
        if 'pileup' in self.services:
            from WMCore.MicroService.MSPileup.MSPileup import MSPileup
            self.msPileup = MSPileup(self.msConfig, logger=self.logger)
            thname = 'MSPileup'
            self.pileupThread = start_new_thread(thname, daemonOpt,
                                                 (self.pileup,
                                                  self.msConfig['interval'],
                                                  self.logger))
            self.logger.info("### Running %s thread %s", thname, self.pileupThread.running())

        # initialize pileup-tasks module
        if 'pileup-tasks' in self.services:
            from WMCore.MicroService.MSPileup.MSPileupTaskManager import MSPileupTaskManager
            self.msPileupTasks = MSPileupTaskManager(self.msConfig, logger=self.logger)
            thname = 'MSPileupTasks'
            self.pileupTasksThread = start_new_thread(thname, daemonOpt,
                                                      (self.pileupTasks,
                                                       self.msConfig['interval'],
                                                       self.logger))
            self.logger.info("### Running %s thread %s", thname, self.pileupTasksThread.running())

        # initialize transferor module
        if 'transferor' in self.services:
            from WMCore.MicroService.MSTransferor.MSTransferor import MSTransferor
            self.msTransferor = MSTransferor(self.msConfig, logger=self.logger)
            thname = 'MSTransferor'
            self.transfThread = start_new_thread(thname, daemon,
                                                 (self.transferor,
                                                  'assigned',
                                                  self.msConfig['interval'],
                                                  self.logger))
            self.logger.info("### Running %s thread %s", thname, self.transfThread.running())

        # initialize monitoring module
        if 'monitor' in self.services:
            from WMCore.MicroService.MSMonitor.MSMonitor import MSMonitor
            self.msMonitor = MSMonitor(self.msConfig, logger=self.logger)
            thname = 'MSMonitor'
            self.monitThread = start_new_thread(thname, daemon,
                                                (self.monitor,
                                                 'staging',
                                                 self.msConfig['interval'],
                                                 self.logger))
            self.logger.info("+++ Running %s thread %s", thname, self.monitThread.running())

        # initialize output module
        if 'output' in self.services:
            from WMCore.MicroService.MSOutput.MSOutput import MSOutput
            reqStatus = ['closed-out', 'announced']

            thname = 'MSOutputConsumer'
            self.msOutputConsumer = MSOutput(self.msConfig, mode=thname, logger=self.logger)
            # set the consumer to run twice faster than the producer
            consumerInterval = self.msConfig['interval'] // 2
            self.outputConsumerThread = start_new_thread(thname, daemon,
                                                         (self.outputConsumer,
                                                          reqStatus,
                                                          consumerInterval,
                                                          self.logger))
            self.logger.info("=== Running %s thread %s", thname, self.outputConsumerThread.running())

            thname = 'MSOutputProducer'
            self.msOutputProducer = MSOutput(self.msConfig, mode=thname, logger=self.logger)
            self.outputProducerThread = start_new_thread(thname, daemon,
                                                         (self.outputProducer,
                                                          reqStatus,
                                                          self.msConfig['interval'],
                                                          self.logger))
            self.logger.info("=== Running %s thread %s", thname, self.outputProducerThread.running())

        # initialize rule cleaner module
        if 'ruleCleaner' in self.services:
            from WMCore.MicroService.MSRuleCleaner.MSRuleCleaner import MSRuleCleaner
            reqStatus = ['announced', 'aborted-completed', 'rejected']
            self.msRuleCleaner = MSRuleCleaner(self.msConfig, logger=self.logger)
            thname = 'MSRuleCleaner'
            self.ruleCleanerThread = start_new_thread(thname, daemon,
                                                      (self.ruleCleaner,
                                                       reqStatus,
                                                       self.msConfig['interval'],
                                                       self.logger))
            self.logger.info("--- Running %s thread %s", thname, self.ruleCleanerThread.running())

        # initialize unmerged module
        if 'unmerged' in self.services:
            from WMCore.MicroService.MSUnmerged.MSUnmerged import MSUnmerged
            self.msUnmerged = MSUnmerged(self.msConfig, logger=self.logger)
            thname = 'MSUnmerged'
            self.unmergedThread = start_new_thread(thname, daemonOpt,
                                                   (self.unmerged,
                                                    self.msConfig['interval'],
                                                    self.logger))
            self.logger.info("--- Running %s thread %s", thname, self.unmergedThread.running())

    def _parseConfig(self, config):
        """
        __parseConfig_
        Parse the MicroService configuration and set any default values.
        :param config: config as defined in the deployment
        """
        self.logger.info("Using the following MicroServices config: %s", config.dictionary_())
        self.services = getattr(config, 'services', [])

        self.msConfig = {}
        self.msConfig.update(config.dictionary_())

        self.msConfig['reqmgrCacheUrl'] = self.msConfig['reqmgr2Url'].replace('reqmgr2',
                                                                              'couchdb/reqmgr_workload_cache')

    def pileup(self, *args, **kwargs):
        """
        MSManager pileup function.
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the pileup thread...")
        res = self.msPileup.status()
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total pileup execution time: %.2f secs", res['execution_time'])
        self.statusPileup = res

    def pileupTasks(self, *args, **kwargs):
        """
        MSManager pileup tasks function.
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the pileup-tasks thread...")
        self.msPileupTasks.executeCycle()
        res = self.msPileupTasks.status()
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total pileup-tasks execution time: %.2f secs", res['execution_time'])
        self.statusPileupTasks = res

    def transferor(self, reqStatus):
        """
        MSManager transferor function.
        It performs input data placement logic for workflows that have
        been assigned in the system. Successful workflows will be moved
        to the staging status in ReqMgr2.
        For references see
        https://github.com/dmwm/WMCore/wiki/ReqMgr2-MicroService-Transferor
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the transferor thread...")
        res = self.msTransferor.execute(reqStatus)
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total transferor execution time: %.2f secs", res['execution_time'])
        self.statusTrans = res

    def monitor(self, reqStatus):
        """
        MSManager monitoring function.
        It performs transfer requests from staging to staged state of ReqMgr2.
        For references see
        https://github.com/dmwm/WMCore/wiki/ReqMgr2-MicroService-Transferor
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the monitor thread...")
        res = self.msMonitor.execute(reqStatus)
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total monitor execution time: %d secs", res['execution_time'])
        self.statusMon = res

    def outputConsumer(self, reqStatus):
        """
        MSManager Output Data Placement function.
        It subscribes the output datasets to the Data Management System.
        For references see
        https://github.com/dmwm/WMCore/wiki/ReqMgr2-MicroService-Output
        reqStatus: Status of requests to work on
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the outputConsumer thread...")
        res = self.msOutputConsumer.execute(reqStatus)
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total outputConsumer execution time: %d secs", res['execution_time'])
        self.statusOutput = res

    def outputProducer(self, reqStatus):
        """
        MSManager MongoDB Uploader function.
        It uploads the documents describing a workflow output Data subscription
        into MongoDb. For references see
        https://github.com/dmwm/WMCore/wiki/ReqMgr2-MicroService-Output
        reqStatus: Status of requests to work on
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the outputProducer thread...")
        res = self.msOutputProducer.execute(reqStatus)
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total outputProducer execution time: %d secs", res['execution_time'])
        self.statusOutput = res

    def ruleCleaner(self, reqStatus):
        """
        MSManager ruleCleaner function.
        It cleans the block level Rucio rules created by WMAgent and
        performs request status transition from ['announced', 'aborted-completed', 'rejected'] to
        '{normal, aborted, rejected}-archived' state of ReqMgr2.
        For references see
        https://github.com/dmwm/WMCore/wiki/ReqMgr2-MicroService-RuleCleaner
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the ruleCleaner thread...")
        res = self.msRuleCleaner.execute(reqStatus)
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total ruleCleaner execution time: %d secs", res['execution_time'])
        self.statusRuleCleaner = res

    def unmerged(self, *args, **kwargs):
        """
        MSManager unmerged function.
        It cleans the Unmerged area of the CMS LFN Namespace
        For references see
        https://github.com/dmwm/WMCore/wiki/ReqMgr2-MicroService-Unmerged
        """
        startTime = datetime.utcnow()
        self.logger.info("Starting the unmerged thread...")
        res = self.msUnmerged.execute()
        endTime = datetime.utcnow()
        self.updateTimeUTC(res, startTime, endTime)
        self.logger.info("Total Unmerged execution time: %d secs", res['execution_time'])
        self.statusUnmerged = res

    def stop(self):
        "Stop MSManager"
        status = None
        # stop MSMonitor thread
        if 'monitor' in self.services and hasattr(self, 'monitThread'):
            self.monitThread.stop()
            status = self.monitThread.running()
        # stop MSPileup thread
        if 'pileup' in self.services and hasattr(self, 'pileupThread'):
            self.pileupThread.stop()  # stop checkStatus thread
            status = self.pileupThread.running()
        # stop MSPileupTasks thread
        if 'pileup-tasks' in self.services and hasattr(self, 'pileupTasksThread'):
            self.pileupTasksThread.stop()  # stop checkStatus thread
            status = self.pileupTasksThread.running()
        # stop MSTransferor thread
        if 'transferor' in self.services and hasattr(self, 'transfThread'):
            self.transfThread.stop()  # stop checkStatus thread
            status = self.transfThread.running()
        # stop MSOutput threads
        if 'output' in self.services and hasattr(self, 'outputConsumerThread'):
            self.outputConsumerThread.stop()
            status = self.outputConsumerThread.running()
        if 'output' in self.services and hasattr(self, 'outputProducerThread'):
            self.outputProducerThread.stop()
            status = self.outputProducerThread.running()
        # stop MSRuleCleaner thread
        if 'ruleCleaner' in self.services and hasattr(self, 'ruleCleanerThread'):
            self.ruleCleanerThread.stop()
            status = self.ruleCleanerThread.running()
        # stop MSUnmerged thread
        if 'unmerged' in self.services and hasattr(self, 'unmergedThread'):
            self.unmergedThread.stop()
            status = self.unmergedThread.running()
        return status

    def info(self, reqName=None, **kwargs):
        """
        Return transfer information for a given request
        :param reqName: request name
        :param kwargs: other arguments that can be provided for different
            microservices. Examples:
            rse: string with the name of the RSE (e.g 'T2_DE_DESY')
        :return: data transfer information for this request
        """
        if 'ruleCleaner' in self.services or 'unmerged' in self.services \
                or 'pileup' in self.services or 'pileup-tasks' in self.services:
            data = {}
        else:
            data = {"request": reqName, "transferDoc": None}
        if reqName:
            transferDoc = None
            # obtain the transfer information for a given request records from couchdb for given request
            if 'monitor' in self.services:
                transferDoc = self.msMonitor.reqmgrAux.getTransferInfo(reqName)
            elif 'transferor' in self.services:
                transferDoc = self.msTransferor.reqmgrAux.getTransferInfo(reqName)
            elif 'output' in self.services:
                transferDoc = self.msOutputProducer.getTransferInfo(reqName)
            if transferDoc:
                # it's always a single document in Couch
                data['transferDoc'] = transferDoc[0]

        if 'unmerged' in self.services:
            detail = strToBool(kwargs.get('detail', False))

            if kwargs.get("rse"):
                data = self.msUnmerged.getStatsFromMongoDB(detail=detail, rse=kwargs['rse'])
        return data

    def get(self, **kwds):
        """
        Fetch data in backend, HTTP GET request
        :param args: input parameters
        :param kwds: positional arguments
        :return: list of results
        """
        if 'pileup' in self.services:
            return self.msPileup.getPileup(**kwds)
        return []

    def post(self, doc):
        """
        Either create or fetch data from backend, HTTP POST request

        :param doc: input JSON doc for HTTP POST request
        :return: list of results
        """
        if 'pileup' in self.services:
            res = []
            if 'pileupName' in doc:
                # this is create POST request
                res = self.msPileup.createPileup(doc)
            if 'query' in doc:
                # this is POST request to get data for a given JSON query
                res = self.msPileup.queryDatabase(doc)
            return res
        return []

    def update(self, doc):
        """
        Update resource in backend, HTTP PUT request

        :param doc: input JSON doc for HTTP PUT request
        :return: list of results
        """
        if 'pileup' in self.services:
            return self.msPileup.updatePileup(doc)
        return []

    def delete(self, spec):
        """
        Delete resource in backend, HTTP DELETE request

        :param doc: input JSON spec for HTTP DELETE request
        :return: list of results
        """
        if 'pileup' in self.services:
            return self.msPileup.deletePileup(spec)
        return []

    def status(self, detail, **kwargs):
        """
        Return the current status of a MicroService and a summary
        of its last execution activity.
        :param detail: boolean used to retrieve some extra information
          regarding the service
        :param kwargs: other arguments that can be provided for different
            microservices. Examples:
            rse_type: string with the type of the RSEs to fetch (MSUnmerged)
                      Meaningful values at the moment are: "t1", "t2t3" and "t2t3us"
        :return: a dictionary
        """
        data = {"status": "OK"}
        if detail and 'transferor' in self.services:
            data.update(self.statusTrans)
        elif detail and 'pileup' in self.services:
            data.update(self.statusPileup)
        elif detail and 'pileup-tasks' in self.services:
            data.update(self.statusPileupTasks)
        elif detail and 'monitor' in self.services:
            data.update(self.statusMon)
        elif detail and 'output' in self.services:
            data.update(self.statusOutput)
        elif detail and 'ruleCleaner' in self.services:
            data.update(self.statusRuleCleaner)
        elif 'unmerged' in self.services:
            if kwargs.get("rse_type"):
                data.update({"rse_type": kwargs.get("rse_type")})
            if detail:
                data.update(self.statusUnmerged)
        return data

    def updateTimeUTC(self, reportDict, startT, endT):
        """
        Given a report summary dictionary and start/end time, update
        the report with human readable timing information
        :param reportDict: summary dictionary
        :param startT: epoch start time for a given service
        :param endT: epoch end time for a given service
        """
        reportDict['start_time'] = startT.strftime("%a, %d %b %Y %H:%M:%S UTC")
        reportDict['end_time'] = endT.strftime("%a, %d %b %Y %H:%M:%S UTC")
        reportDict['execution_time'] = (endT - startT).total_seconds()



#!/usr/bin/env python

"""
Auto generated stub be careful with editing,
Inheritance is preferred.
"""
from __future__ import absolute_import

import logging
import threading

from WMCore.Agent.Harness import Harness
from WMQuality.Emulators.RucioClient.MockRucioApi import SITES as DUMMY_SITES
from .WMAgentTasks import WMAgentTasks


class WMAgentEmulator(Harness):
    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)

    def preInitialization(self):
        resources = self.populateResources()
        # Add event loop to worker manager
        myThread = threading.currentThread()
        pollInterval = 1
        logging.info("Setting poll interval to %s seconds" % pollInterval)
        myThread.workerThreadManager.addWorker(WMAgentTasks(resources), pollInterval)

        return

    def populateResources(self):
        """
        emulating resource db which can represent
        {site: job} format
        """
        jobSites = {}
        for site in DUMMY_SITES:
            jobSites[site] = 100
        return jobSites



#!/usr/bin/env python
"""
WorkQueue Start Policy

"""
from copy import deepcopy
from WMCore.WMFactory import WMFactory

startFac = WMFactory(__name__, __name__)


def startPolicy(name, startMap, rucioObj, logger=None):
    """Load a start policy"""
    # Take splitting policy from workload & load specific policy for this queue
    # Workload may also directly specify specific splitter implementation
    if name in startMap:
        # take mapping from queue
        policyName = startMap[name]['name']
    else:
        # workload directly specifies splitting algo
        policyName = name
    try:
        args = deepcopy(startMap[name]['args'])
    except IndexError:
        args = {}
    args['rucioObject'] = rucioObj
    args['logger'] = logger
    return startFac.loadObject(policyName,
                               args,
                               storeInCache=False,
                               )


__all__ = []



#!/usr/bin/env python
"""
_GetOpenBlocks_

Oracle implementation of DBS3Buffer.GetOpenBlocks
"""

from WMComponent.DBS3Buffer.MySQL.GetOpenBlocks import GetOpenBlocks as MySQLGetOpenBlocks

class GetOpenBlocks(MySQLGetOpenBlocks):
    """
    Identical to MySQL version

    """



#!/usr/bin/env python

"""
_Queries_

This module implements the mysql backend for the persistent threadpool.

"""

import threading

from WMCore.Database.DBFormatter import DBFormatter


class Queries(DBFormatter):
    """
    _Queries_

    This module implements the mysql backend for the persistent threadpool.

    """

    def __init__(self):
        myThread = threading.currentThread()
        DBFormatter.__init__(self, myThread.logger, myThread.dbi)

    def selectWork(self, args, pooltable='tp_threadpool'):
        """
        Select work that is not yet being processed.
        """
        # this query takes place in a locked section so
        # we do not have to worry about multiple slaves
        # getting the same work.
        result = ''
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
                         'tp_threadpool_buffer_out']:
            sqlStr = """
SELECT min(id) FROM %s WHERE component = :component AND
thread_pool_id = :thread_pool_id AND state='queued'
        """ % (pooltable)
            result = self.execute(sqlStr, args)
        else:
            sqlStr = """
SELECT min(id) FROM %s WHERE state='queued'
        """ % (pooltable)
            result = self.execute(sqlStr, {})
        return self.formatOne(result)

    def retrieveWork(self, args, pooltable='tp_threadpool'):
        """
        Retrieve work (its payload) from a queue.
        """

        sqlStr = """
SELECT id, event,payload FROM %s WHERE id = :id
        """ % (pooltable)
        result = self.execute(sqlStr, args)
        return self.formatOne(result)

    def tagWork(self, args, pooltable='tp_threadpool'):
        """
        Tag work in queue as being processed.
        """
        sqlStr = """
UPDATE %s SET state='process' WHERE id = :id
        """ % (pooltable)
        self.execute(sqlStr, args)

    def removeWork(self, args, pooltable='tp_threadpool'):
        """
        Remove work from the queue.
        """
        sqlStr = """
DELETE FROM %s WHERE id = :id
        """ % (pooltable)
        self.execute(sqlStr, args)

    def updateWorkStatus(self, args, pooltable='tp_threadpool'):
        """
        Updates work status of work being processed.
        """
        # differentiate between one queue and multi queue
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
                         'tp_threadpool_buffer_out']:
            sqlStr = """UPDATE %s SET state = 'queued'
                          WHERE component = :componentName AND
                                thread_pool_id = :thread_pool_id
                     """ % (pooltable)
            self.execute(sqlStr, args)
        else:
            sqlStr = "UPDATE %s SET state = 'queued'" % (pooltable)
            self.execute(sqlStr, {})

        return

    def getQueueLength(self, args, pooltable='tp_threadpool'):
        """
        Returns the queue length from the different tables
        through a parameterized query.
        """
        # differentiate between onequeu and multi queue
        result = None
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
                         'tp_threadpool_buffer_out']:
            sqlStr = """
SELECT COUNT(*) FROM  %s WHERE component = :componentName
AND thread_pool_id = :thread_pool_id
        """ % (pooltable)
            result = self.execute(sqlStr, args)
        else:
            sqlStr = """
SELECT COUNT(*) FROM  %s
        """ % (pooltable)
            result = self.execute(sqlStr, {})
        return self.formatOne(result)[0]

    def moveWorkFromBufferIn(self, source, target):
        """
        Moves work from buffer in to main queue or buffer out
        """

        sqlStr1 = ''
        sqlStr2 = ''
        if source == 'tp_threadpool_buffer_in':
            sqlStr1 = """
INSERT INTO %s(event,component,payload,thread_pool_id) SELECT event,component,payload,thread_pool_id FROM %s FOR UPDATE
            """ % (target, source)
            sqlStr2 = """ DELETE FROM %s """ % (source)
        else:
            sqlStr1 = """
INSERT INTO %s(event,payload) SELECT event,payload FROM %s
            """ % (target, source)
            sqlStr2 = """ DELETE FROM %s """ % (source)
        self.execute(sqlStr1, {})
        self.execute(sqlStr2, {})

    def moveWorkToBufferOut(self, args, source, target, limit):
        """
        Moves work from buffer in or main queue to the buffer out
        table.
        """

        sqlStr1 = ''
        sqlStr2 = ''
        if source in ['tp_threadpool', 'tp_threadpool_buffer_in', \
                      'tp_threadpool_buffer_out']:
            # we need a for update in the select to prevent (harmless) deadlock
            # situations with innodb
            sqlStr1 = """
INSERT INTO %s(event,component,payload,thread_pool_id)
SELECT event,component,payload,thread_pool_id FROM %s
WHERE component= :component AND thread_pool_id = :thread_pool_id
ORDER BY id LIMIT %s FOR UPDATE
            """ % (target, source, limit)
            sqlStr2 = """
DELETE FROM %s
WHERE component= :component AND thread_pool_id = :thread_pool_id
ORDER BY id LIMIT %s""" % (source, limit)
            self.execute(sqlStr1, args)
            self.execute(sqlStr2, args)
        else:
            sqlStr1 = """
INSERT INTO %s(event,payload) SELECT event,payload FROM %s ORDER BY id LIMIT %s
            """ % (target, source, limit)
            sqlStr2 = """
DELETE FROM %s ORDER BY id LIMIT %s
            """ % (source, limit)
            self.execute(sqlStr1, {})
            self.execute(sqlStr2, {})

    def insertWork(self, args, pooltable='tp_threadpool'):
        """
        Inserts work into the database in case no thread can be found.
        """
        # differentiate between onequeu and multi queue
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
                         'tp_threadpool_buffer_out']:
            sqlStr = """
INSERT INTO %s(event,component,payload,thread_pool_id)
VALUES(:event,:component,:payload,:thread_pool_id)
        """ % (pooltable)
            self.execute(sqlStr, args)
        else:
            sqlStr = """
INSERT INTO %s(event,payload)
VALUES(:event,:payload)
        """ % (pooltable)
            self.execute(sqlStr, {'event': args['event'], \
                                  'payload': args['payload']})

    def insertThreadPoolTables(self, threadpool):
        """
        __insertThreadPoolTable

        Inserts tables for threadpool (one for each threadpool)
        when multi queue is enabled.
        """
        sqlStr = """
CREATE TABLE %s(
   id                      int(11)      NOT NULL auto_increment,
   event                   varchar(255) NOT NULL,
   payload                 text         NOT NULL,
   state                 enum('queued','process') default 'queued',
   primary key(id)
   ) ENGINE=InnoDB ROW_FORMAT=DYNAMIC""" % (threadpool)

        self.execute(sqlStr, {})

    def execute(self, sqlStr, args):
        """"
        __execute__
        Executes the queries by getting the current transaction
        and dbinterface object that is stored in the reserved words of
        the thread it operates in.
        """
        # FIXME: we use this method in all kinds of places perhaps upgrade
        # FIXME: this method?
        myThread = threading.currentThread()
        currentTransaction = myThread.transaction
        return currentTransaction.processData(sqlStr, args)



#!/usr/bin/env python
"""
_Diagnostic_


Interface and base class for step specific diagnostic handlers


"""
from builtins import object

from WMCore.FwkJobReport.Report import FwkJobReportException


class DiagnosticHandler(object):
    """
    _DiagnosticHandler_

    Interface definition for handlers for a specific error condition


    """
    def __call__(self, errorCode, executorInstance, **args):
        """
        _operator(errCode, executor)_

        Override to act on a particular error, use the executorInstance
        to access things like the step, logfiles, and report.

        Args will be used to provide extra information such as Exception
        instances etc

        """
        msg = "DiagnosticHandler.__call__ not "
        msg += "implemented for class %s" % self.__class__.__name__
        raise NotImplementedError(msg)

    def parse(self, executorInstance, jobRepXml):
        """
        Add an error to report if parsing the xml fails.
        """
        try:
            executorInstance.report.parse(jobRepXml, executorInstance.stepName)
        except FwkJobReportException as ex:
            # Job report is bad, the parse already puts a 50115 in the file
            msg = "Error reading XML job report file, possibly corrupt XML File:\n"
            msg += "Details: %s" % str(ex)
            executorInstance.report.addError(executorInstance.stepName, 50115, "BadFWJRXML", msg)
            raise


class DefaultDiagnosticHandler(DiagnosticHandler):
    """
    _DefaultDiagnosticHandler_

    Catch-all that just adds information to the report

    """
    def __call__(self, errorCode, executorInstance, **args):
        pass





class Diagnostic(object):
    """
    _Diagnostic_

    Base class for a Diagnostic implementation specific to a step type
    Also works as a bare minimum Diagnostic if overriding is not needed

    """
    def __init__(self):
        self.handlers = {}
        self.defaultHandler = DefaultDiagnosticHandler()



    def __call__(self, errCode, executor, **args):
        """
        _operator(errCode, executor, args)_

        Invoke the diagnostic to produce an error report

        """
        handler = self.handlers.get(errCode, self.defaultHandler)
        handler(errCode, executor, **args)
        executor.saveReport()



#!/usr/bin/env python
"""
_LoadFromUID_

MySQL implementation of JobGroup.LoadFromUID
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class LoadFromUID(DBFormatter):
    sql = """SELECT id, subscription, guid, output, last_update
             FROM wmbs_jobgroup WHERE guid = :guid"""

    def formatDict(self, result):
        """
        _formatDict_

        Cast the id, subscription, output and last_update attributes to integers
        because formatDict() turns everything into strings.
        """
        formattedResult = DBFormatter.formatDict(self, result)[0]
        formattedResult["id"]           = int(formattedResult["id"])
        formattedResult["subscription"] = int(formattedResult["subscription"])
        formattedResult["output"]       = int(formattedResult["output"])
        formattedResult["last_update"]  = int(formattedResult["last_update"])
        formattedResult["uid"]          = formattedResult["guid"]
        del formattedResult['guid']
        return formattedResult

    def execute(self, uid, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"guid": uid}, conn = conn,
                                      transaction = transaction)
        return self.formatDict(result)



#/usr/bin/env python2.4
"""
_Destroy_

"""




import threading

from WMCore.Database.DBCreator import DBCreator

from WMCore.WMException import WMException
from WMCore.WMExceptions import WMEXCEPTION

class Destroy(DBCreator):
    def __init__(self, logger = None, dbi = None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """
        myThread = threading.currentThread()

        if logger == None:
            logger = myThread.logger
        if dbi == None:
            dbi = myThread.dbi

        DBCreator.__init__(self, logger, dbi)

        self.create = {}
        self.delete = {}
        self.delete["03tp_threadpool"]             = "DROP TABLE tp_threadpool"
        self.delete["02tp_threadpool_buffer_in"]   = "DROP TABLE tp_threadpool_buffer_in"
        self.delete["01tp_threadpool_buffer_out"]  = "DROP TABLE tp_threadpool_buffer_out"



'''
Created on Feb 15, 2010

'''
from WMCore.Configuration import Configuration
from WMCore.WMBase import getWMBASE
import os.path

config = Configuration()
config.section_("Emulator")
config.Emulator.PhEDEx = True
config.Emulator.DBSReader = True
config.Emulator.ReqMgr = True



#!/usr/bin/env python
# pylint: disable=E1101, W0212
"""
_ConfigSectionTree_

Extension for a normal ConfigSection to provide a Tree structure
of ConfigSections

"""



from builtins import str, bytes, object

from WMCore.Configuration import ConfigSection


def nodeName(node):
    """
    _nodeName_

    Util for extracting node name

    """
    return node._internal_name

def nodeParent(node):
    """
    _nodeParent_

    Util for extracting Node Parent reference

    """
    return node._internal_parent_ref

def listNodes(topNode):
    """
    _listNodes_

    Util to return a list of all node names in a NodeData structure
    returns them in execution order

    """
    result = []
    result.append(nodeName(topNode))
    for child in topNode.tree.childNames:
        result.extend(listNodes(getattr(topNode.tree.children, child)))
    return result

def listChildNodes(topNode):
    """
    _listChildNodes_

    ListNodes but without including the top node
    """
    result = []
    for child in topNode.tree.childNames:
        result.extend(listNodes(getattr(topNode.tree.children, child)))
    return result

def listFirstGenChildNodes(topNode):
    """
    _listFirstGenChildNodes_

    Return a list of the first generator child nodes.
    """
    # no real need to sort it, but we better have the same order between Py2/Py3
    return sorted(list(topNode.tree.childNames))

def nodeMap(node):
    """
    _nodeMap_

    Generate a map of node name to node data instance
    Note: This will *not* preserve the execution order of the child nodes
    and should not be used to traverse and operate on nodes where order matters

    """
    result = {}
    result[nodeName(node)] = node
    for child in node.tree.childNames:
        result.update(nodeMap(getattr(node.tree.children, child)))
    return result


def findTopNode(node):
    """
    _findTopNode_

    Find the top node of a tree of nodes given an arbitrary node in the tree
    Checks for non None parent, will also stop if the internal_treetop flag
    is set for the parent
    """
    parent = nodeParent(node)
    if parent == None:
        return node
    if getattr(node, "_internal_treetop", False):
        return node

    if getattr(parent, "_internal_treetop", False):
        return parent
    return findTopNode(nodeParent(node))


def allNodeNames(node):
    """
    _allNodeNames_

    Find the top node in the tree and then get a list of all known names

    """
    topNode = findTopNode(node)
    return listNodes(topNode)



def addNode(currentNode, newNode):
    """
    _addNode_

    Add a child node to the current node provided

    """
    newName = nodeName(newNode)
    allNames = allNodeNames(currentNode)

    if newName in allNames:
        msg = "Duplicate Node Name %s already exists in tree\n" % newName
        msg += "%s\n" % allNames
        raise RuntimeError(msg)

    setattr(currentNode.tree.children, newName, newNode)
    currentNode.tree.childNames.append(newName)
    newNode.tree.parent = nodeName(currentNode)
    return

def addTopNode(currentNode, newNode):
    """
    _addTopNode_

    Add a child node to the current node provided but insert it
    at the head of the childNames list.
    """
    newName = nodeName(newNode)
    allNames = allNodeNames(currentNode)

    if newName in allNames:
        msg = "Duplicate Node Name %s already exists in tree\n" % newName
        msg += "%s\n" % allNames
        raise RuntimeError(msg)

    setattr(currentNode.tree.children, newName, newNode)
    currentNode.tree.childNames.insert(0, newName)
    newNode.tree.parent = nodeName(currentNode)
    return

def deleteNode(topNode, childName):
    """
    _deleteNode_

    Given a node within the tree, delete the child
    with the given name if it exists
    """
    if hasattr(topNode.tree.children, childName):
        delattr(topNode.tree.children, childName)
        topNode.tree.childNames.remove(childName)

def getNode(node, nodeNameToGet):
    """
    _getNode_

    Given a node within the tree, find the node with the name provided &
    return it.
    returns None if not found

    """
    topNode = findTopNode(node)
    mapping = nodeMap(topNode)
    return mapping.get(nodeNameToGet, None)

def findTop(node):
    """
    _findTop_

    Ignoring tree structure, find the top node that contains the node
    provided.

    Will work for any ConfigSection, not limited to ConfigSectionTree

    """
    if node._internal_parent_ref == None:
        return node
    return findTop(node._internal_parent_ref)


def nodeIterator(node):
    """
    _nodeIterator_

    Generator function that delivers all nodes in order

    """
    for i in listNodes(node):
        yield getNode(node, i)

def nodeChildIterator(node):
    """
    _nodeChildeIterator_

    iterate over all nodes in order, except for the top node passed to this method
    """
    for i in listChildNodes(node):
        yield getNode(node, i)

def firstGenNodeChildIterator(node):
    """
    _firstGenNodeChildIterator_

    Iterator over all the first generation child nodes.
    """
    for i in listFirstGenChildNodes(node):
        yield getNode(node, i)

def format(value):
    """
    _format_

    format a value as python
    keep parameters simple, trust python...
    """
    if isinstance(value, (str, bytes)):
        value = "\'%s\'" % value
    return str(value)

def formatNative(value):
    """
    _formatNative_

    Like the format function, but allowing passing of ints, floats, etc.
    """

    if isinstance(value, int):
        return value
    if isinstance(value, float):
        return value
    if isinstance(value, list):
        return value
    if isinstance(value, dict):
        return value
    else:
        return format(value)

class TreeHelper(object):
    """
    _TreeHelper_

    Convienience wrapper for a ConfigSectionTree that provides
    all the util methods as a wrapper class to avoid method name
    and attribute collisions in the underlying ConfigSection

    """
    def __init__(self, cfgSectTree):
        self.data = cfgSectTree

    def name(self):
        """get name of this node"""
        return nodeName(self.data)

    def setTopOfTree(self):
        """
        flag this node as the top of the tree
        """
        self.data._internal_treetop = True

    def isTopOfTree(self):
        """
        _isTopOfTree_

        Determine if this section is the top of the tree.
        """
        return self.data._internal_treetop

    def listNodes(self):
        """list this node and all subnodes"""
        return listNodes(self.data)

    def getNode(self, nodeNameToGet):
        """get a node by name from this tree"""
        return getNode(self.data, nodeNameToGet)

    def getNodeWithHelper(self, nodeNameToGet):
        """get a node wrapped in a TreeHelper instance"""
        return TreeHelper(getNode(self.data, nodeNameToGet))

    def addNode(self, newNode):
        """
        add a new Node, newNode can be a ConfigSectionTree or a TreeHelper
        """
        if isinstance(newNode, TreeHelper):
            return addNode(self.data, newNode.data)
        return addNode(self.data, newNode)

    def addTopNode(self, newNode):
        """
        _addTopNode_

        Add a child node to the current node provided but insert it
        at the head of the childNames list.
        """
        if isinstance(newNode, TreeHelper):
            return addTopNode(self.data, newNode.data)
        return addTopNode(self.data, newNode)

    def deleteNode(self, nodeName):
        """
        _deleteNode

        Delete a child node given its name,
        if it doesn't exists then do nothing
        """
        deleteNode(self.data, nodeName)
        return


    def allNodeNames(self):
        """get list of all known node names in the tree containing this node"""
        return allNodeNames(self.data)

    def findTopNode(self):
        """get ref to the top node in the tree containing this node"""
        return findTopNode(self.data)

    def getTopConfigSection(self):
        """
        _getTopConfigSection_

        Ignore tree structure & fetch the absolute top of the pile
        ConfigSection containing this node

        """
        return findTop(self.data)

    def nodeIterator(self):
        """
        generator for processing all subnodes in execution order
        """
        for i in listNodes(self.data):
            yield getNode(self.data, i)

    def nodeChildIterator(self):
        """
        generator for processing all subnodes in execution order
        """
        for i in listChildNodes(self.data):
            yield getNode(self.data, i)

    def firstGenNodeChildIterator(self):
        """
        _fristGenNodeChildIterator_

        Iterate over all the first generation child nodes.
        """
        for i in listFirstGenChildNodes(self.data):
            yield getNode(self.data, i)

    def pythoniseDict(self, **options):
        """
        convert self into dict of python format strings with
        values in value position

        """
        prefix     = options.get('prefix',   None)
        sections   = options.get('sections',   False)

        if prefix != None:
            myName = "%s.%s" % (prefix, self.data._internal_name)
        else:
            myName = self.data._internal_name

        result = {}

        for attr in self.data._internal_settings:
            if attr in self.data._internal_children:
                if sections:
                    result["%s.section_(\'%s\')" % (myName, attr)] = '_Section_'
                result.update(TreeHelper(getattr(self.data, attr)).pythoniseDict(prefix = myName))
                continue

            #This is potentially dangerous, because it adds lists, dicts.
            result["%s.%s" %(myName, attr)] = formatNative(getattr(self.data, attr))

        return result


    def addValue(self, value):
        """
        _addValue_

        adds an arbitrary value as a dictionary.  Can have multiple values
        """

        if not isinstance(value, dict):
            raise Exception("TreeHelper.addValue passed a value that was not a dictionary")

        for key in value:
            splitList = key.split('.')
            setResult = value[key]
            if len(splitList) == 1:
                #Then there is only one level, and we put it here
                setattr(self.data, key, setResult)
            else:
                if splitList[0] in self.data.listSections_():
                    #If the section exists, go to it directly
                    helper = TreeHelper(getattr(self.data, splitList[0]))
                    helper.addValue({splitList[1:]:setResult})
                else:
                    #If the section doesn't exist, create it
                    self.data.section_(splitList[0])
                    helper = TreeHelper(getattr(self.data, splitList[0]))
                    helper.addValue({"".join(splitList[1:]):setResult})
        return





class ConfigSectionTree(ConfigSection):
    """
    _ConfigSectionTree_

    Node Tree structure that can be embedded into a ConfigSection
    structure

    """

    def __init__(self, name):
        ConfigSection.__init__(self, name)
        self._internal_treetop = False
        self.section_("tree")
        self.tree.section_("children")
        self.tree.childNames = []
        self.tree.parent = None



#!/usr/bin/env python
"""
_RemoveDuplicates_

MySQL implementation of Workflow.RemoveDuplicates
"""

from future.utils import listvalues

from WMCore.Database.DBFormatter import DBFormatter

class RemoveDuplicates(DBFormatter):
    sql = """SELECT lfn FROM wmbs_file_details
               INNER JOIN wmbs_fileset_files ON
                 wmbs_file_details.id = wmbs_fileset_files.fileid
               INNER JOIN wmbs_subscription ON
                 wmbs_fileset_files.fileset = wmbs_subscription.fileset
               INNER JOIN wmbs_workflow ON
                 wmbs_subscription.workflow = wmbs_workflow.id
             WHERE wmbs_file_details.lfn = :lfn AND
                   wmbs_workflow.name = :workflow"""

    def execute(self, files, workflow, conn = None, transaction = False):
        binds = []
        newFiles = {}
        for file in files:
            binds.append({"lfn": file["lfn"], "workflow": workflow})
            newFiles[file["lfn"]] = file

        results = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        for result in self.formatDict(results):
            if result["lfn"] in list(newFiles):
                del newFiles[result["lfn"]]

        return listvalues(newFiles)



#!/usr/bin/env python
"""
Fileset.py

Created by Dave Evans on 2010-03-18.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from WMCore.DataStructs.WMObject import WMObject


class Fileset(dict, WMObject):
    """
    _Fileset_

    """

    def __init__(self, **options):
        dict.__init__(self)
        WMObject.__init__(self)
        self.setdefault("name", None)
        self.setdefault("files", {})
        self.update(options)
        self.collection = None

    def setCollection(self, collection):
        """
        _setCollection_

        Associate this fileset with a collection.
        """
        self.collectionName = collection["name"]
        self.collectionType = collection["type"]
        return

    def create(self):
        """
        _create_

        create a new fileset within a collection
        """
        pass

    def populate(self):
        """
        _populate_

        populate information about this fileset
        """
        pass

    def drop(self):
        """
        _drop_

        Remove the fileset from its collection
        """
        pass

    def add(self, files, mask):
        """
        _add_

        Add files to this fileset
        files should be a list of WMCore.DataStruct.File objects

        """
        pass

    def listFiles(self):
        """
        _listFiles_

        Iterate/yield the files in this fileset
        """
        pass

    def fileCount(self):
        """
        _fileCount_

        Total number of files in this fileset
        """
        pass

    def fileset(self):
        """
        _fileset_

        Create and return an instance of a WMCore.DataStructs.Fileset
        containing the files in this (ACDC) fileset

        """
        pass



from __future__ import print_function
import unittest
import cherrypy
import logging
import threading
import os

#decorator import for RESTServer setup
from WMQuality.REST.ServerSetup import RESTMainTestServer
import WMQuality.CherrypyTestInit as CherrypyTestInit


class RESTBaseUnitTestWithDBBackend(unittest.TestCase):

    def setUp(self, initRoot = True):
        """
            unittest inherits this class
            should have set
            self.setConfig(config)
            config is WMCore.REST application config

            i.e.
            class ClildClass(RESTBaseUnitTestWithDBBackend):

                def setUp(self):
                    self.setConfig(WMCore.ReqMgr.Config)
                    # following setters are optional
                    self.setCouchDBs([("reqmgr_workload_config", "ReqMgr")])
                    self.setSchemaModules(["WMCore.WMBS"])


        """
        if self.schemaModules or self.couchDBs:
            from WMQuality.TestInitCouchApp import TestInitCouchApp
            self.testInit = TestInitCouchApp(__file__)
            self.testInit.setLogging() # logLevel = logging.SQLDEBUG

            if self.schemaModules:
                self.testInit.setDatabaseConnection()
                self.testInit.setSchema(customModules = self.schemaModules,
                                        useDefault = False)
                # Now pull the dbURL from the factory
                # I prefer this method because the factory has better error handling
                # Also because then you know everything is the same
                myThread = threading.currentThread()
                self.config.setDBUrl(myThread.dbFactory.dburl)

            if self.couchDBs:
                for (dbName, couchApp) in self.couchDBs:
                    if couchApp:
                        self.testInit.setupCouch(dbName, couchApp)
                    else:
                        self.testInit.setupCouch(dbName)


        logging.info("RESTBaseUnitTestWithDBBackend configuration: %s" % self.config)

        self.initRoot = initRoot
        if initRoot:
            self.server = RESTMainTestServer(self.config, os.getcwd(), self._testMethodName)
            CherrypyTestInit.start(self.server)
            self.jsonSender = self.server.jsonSender
            # find the way to check the api with the permission
            self.test_authz_key = self.server.test_authz_key
            print("init root")

    def tearDown(self):
        logging.info("RESTBaseUnitTestWithDBBackend executing tearDown")
        if self.initRoot:
            CherrypyTestInit.stop(self.server)
            self.test_authz_key = None

        if self.schemaModules:
            self.testInit.clearDatabase()

        if self.couchDBs:
            self.testInit.tearDownCouch()

        self.config = None
        self.jsonSender = None
        return

    def setSchemaModules(self, schemaModules):
        """
        This need to be set if backend db connection is needed
        ie.
        schemaModules = ["WMCore.WMBS","WMComponent.DBS3Buffer","WMCore.BossAir"]
        """
        self.schemaModules = schemaModules or []

    def setCouchDBs(self, couchDBs):
        """
        This need to be set if counchdb connection is needed
        couchDBs = [("reqmgr_workload_config", "ReqMgr"),  ]
        """
        self.couchDBs = couchDBs or []

    def setConfig(self, config):
        self.config = config



#!/usr/bin/env python
"""
_AddBulkParentage_

Oracle implementation of Files.AddBulkParentage
"""




from WMCore.WMBS.MySQL.Files.AddBulkParentage import AddBulkParentage as ParentageMySQL

class AddBulkParentage(ParentageMySQL):
    pass



"""
Module with validation functions for the Auxiliary-based
RESTful APIs, such as unifiedconfig, wmagentconfig, etc.
"""


def validateUnifiedConfig(data):
    """
    Function to validate content of the Unified configuration
    before it gets persisted in the database.
    :param data: dictionary with the Unified configuration
    :return: raises an exception in case of problems, otherwise None
    """
    SUPPORTED_KEYS = {"tiers_to_DDM", "tiers_no_DDM", "tiers_with_no_custodial"}
    NESTED_KEYS = {"value", "description"}
    baseError = "Unified schema error."
    if not isinstance(data, dict):
        raise RuntimeError(f"{baseError} It needs to be a dictionary, not: {type(data)}")

    # does user data contains all of the required top level parameters?
    if not SUPPORTED_KEYS.issubset(set(data.keys())):
        raise RuntimeError(f"{baseError} It is missing some of the required parameters: {SUPPORTED_KEYS}")

    for keyAttr in data:
        # is the user data parameter listed as a supported parameter?
        if keyAttr not in SUPPORTED_KEYS:
            raise RuntimeError(f"{baseError} It has unsupported parameter: {keyAttr}")

        # does user data contains all of the internal keys within each top level parameter?
        if not NESTED_KEYS.issubset(set(data.get(keyAttr, {}).keys())):
            # are all elements of expected inner keys in data[keyAttr]?
            msg = f"{baseError} Attribute '{keyAttr}' is missing "
            msg += f"some of the nested required keys: {NESTED_KEYS}"
            raise RuntimeError(msg)

        # are the nested keys in the user data parameter provided supported in the schema?
        for nestedAttr, nestedValue in data[keyAttr].items():
            if nestedAttr not in NESTED_KEYS:
                msg = f"{baseError} Attribute '{keyAttr}' has "
                msg += f"unsupported nested key: {nestedAttr}"
                raise RuntimeError(msg)
            # then validate the value data type as well
            if nestedAttr == "value":
                if not isinstance(nestedValue, list):
                    msg = f"{baseError} Data type of {keyAttr}.{nestedAttr} "
                    msg =+ f"should be 'list', not '{type(nestedValue)}'"
                    raise RuntimeError(msg)
            elif nestedAttr == "description":
                if not isinstance(nestedValue, str):
                    msg = f"{baseError} Data type of {keyAttr}.{nestedAttr} "
                    msg =+ f"should be 'str', not '{type(nestedValue)}'"
                    raise RuntimeError(msg)
    # if it got here, then everything is fine!



#!/usr/bin/env python
"""
File       : MSPileupReport.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSPileup report module
"""

# WMCore modules
from Utils.Timers import gmtimeSeconds, encodeTimestamp


class MSPileupReport():
    """
    MSPileupReport class represents MSPileup report object(s)
    """
    def __init__(self, autoExpire=3600, autoCleanup=False):
        """
        Constructor for MSPileup object
        """
        self.docs = []
        self.autoExpire = autoExpire
        self.autoCleanup = autoCleanup

    def addEntry(self, task, uuid, entry):
        """
        Add new entry into MSPileup documents

        :param task: task name
        :param uuid: unique id of the entry
        :param entry: entry message or any other object to store
        """
        if self.autoCleanup:
            self.purgeExpired()
        gmtime = gmtimeSeconds()
        report = {'gmtime': gmtime, 'uuid': uuid,
                  'timestamp': encodeTimestamp(gmtime),
                  'entry': entry, 'task': task}
        self.docs.append(report)

    def purgeExpired(self):
        """
        Purge expired records from internal docs
        """
        gmtime = gmtimeSeconds()
        for entry in list(self.docs):
            if gmtime - entry['gmtime'] > self.autoExpire:
                self.docs.remove(entry)

    def getDocuments(self):
        """
        Return report documents
        """
        if self.autoCleanup:
            self.purgeExpired()
        return self.docs

    def getReportByUuid(self):
        """
        Return report documents in dictonary form with uuid's as keys
        """
        if self.autoCleanup:
            self.purgeExpired()
        rdict = {}
        for doc in self.docs:
            uuid = doc['uuid']
            timestamp = doc['timestamp']
            entry = doc['entry']
            task = doc['task']
            record = f"{timestamp} {task} task {uuid} {entry}"
            rdict.setdefault(uuid, []).append(record)
        return rdict



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : utils.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description:
"""
from __future__ import print_function
from builtins import str as newstr, bytes, int
from future.utils import viewitems

from future import standard_library

from Utils.PythonVersion import PY3

standard_library.install_aliases()

# system modules
import cgi
import json
import time
import hashlib
import cherrypy
from urllib.error import URLError

# WMCore Modules
from Utils.Utilities import encodeUnicodeToBytes, encodeUnicodeToBytesConditional


def tstamp():
    "Generic time stamp"
    return time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())

def gen_color(val):
    "Generate unique color code for given string value"
    keyhash = hashlib.md5()
    keyhash.update(encodeUnicodeToBytesConditional(val, condition=PY3))
    col = '#%s' % keyhash.hexdigest()[:6]
    return col

def quote(data):
    """
    Sanitize the data using cgi.escape.
    """
    if  isinstance(data, (int, float, dict, list)):
        res = data
    else:
        try:
            if  data:
                res = cgi.escape(data, quote=True)
            else:
                res = ""
        except Exception as exc:
            print("Unable to cgi.escape(%s, quote=True)" % data)
            res = ""
    return res

def json2form(jsondata, indent=2, keep_first_value=True):
    "Convert input json dict into one used by HTML form"
    if  keep_first_value:
        for key, val in viewitems(jsondata):
            if  isinstance(val, list):
                if len(val) == 0:
                    jsondata[key] = ""
                else:
                    jsondata[key] = val[0]
    return json.dumps(jsondata, indent=2)

def json2table(jsondata, web_ui_map, visible_attrs=None, selected={}):
    """
    Convert input json dict into HTML table based on assumtion that
    input json is in a simple key:value form.
    """
    table = """<table class="table-bordered width-100">\n"""
    table += "<thead><tr><th>Field</th><th>Value</th></tr></thead>\n"
    keys = sorted(jsondata.keys())
    # move up keys whose values have REPLACE prefix
    priority_keys = []
    rest_keys = []
    for key in keys:
        val = jsondata[key]
        if  isinstance(val, (newstr, bytes)) and val.startswith('REPLACE-'):
            priority_keys.append(key)
        else:
            rest_keys.append(key)
    cells = {}
    for key in priority_keys+rest_keys:
        val = jsondata[key]
        if  isinstance(val, list) and not val: # empty list replace with input text tag
            val = ""
        if  isinstance(val, list):
            if  not visible_attrs:
                sel = '<textarea name="%s" class="width-100">%s</textarea>' \
                        % (key, json.dumps(val))
            else:

                MULTI_SELECTION_KEYS = ['SiteWhitelist', 'SiteBlacklist']
                if key in MULTI_SELECTION_KEYS:
                    sel = "<select class=\"width-100\" name=\"%s\" multiple>" % key
                else:
                    sel = "<select class=\"width-100\" name=\"%s\">" % key

                if key in selected:
                    values = val
                else:
                    values = sorted(val)

                if  key in ['CMSSWVersion', 'ScramArch']:
                    values.reverse()
                # when there is no value to be selected
                if key in selected and not selected[key]:
                    sel += "<option selected disabled>--select an option--</option>"
                for item in values:
                    if key in selected and item in selected[key]:
                        sel += "<option value=\"%s\" selected=\"selected\">%s</option>" % (item, item)
                    else:
                        sel += "<option value=\"%s\">%s</option>" % (item, item)
                sel += "</select>"
            val = sel
        elif isinstance(val, (newstr, bytes)):
            if  val.startswith('REPLACE-'):
                val = '<input type="text" name="%s" placeholder="%s" class="width-100">'\
                        % (key, val)
            elif  len(val) < 80:
                val = '<input type="text" name="%s" value="%s" class="width-100" />' % (key, val)
            else:
                val = '<textarea name="%s" class="width-100">%s</textarea>' % (key, val)
        elif isinstance(val, (dict, list)):
            val = '<textarea name="%s" class="width-100">%s</textarea>' % (key, json.dumps(val))
        else:
            val = '<input type="text" name="%s" value="%s" class="width-100" />' % (key, val)
        if  key in web_ui_map:
            kname = web_ui_map[key]
        else:
            # use original key
            kname = key
        cells[key] = (kname, val)
    if  visible_attrs and isinstance(visible_attrs, list):
        for attr in visible_attrs:
            key, val = cells.pop(attr)
            if  key in web_ui_map:
                kname = web_ui_map[key]
            else:
                # use original key
                kname = key
            val = val.replace('width-100', 'width-100 visible_input')
            table += "<tr><td>%s</td><td class=\"visible\">%s</td></tr>\n" % (kname, val)
    for key, pair in viewitems(cells):
        kname, val = pair
        if  not visible_attrs:
            val = val.replace('<input', '<input readonly')
            val = val.replace('<textarea', '<textarea readonly')
            val = val.replace('<select', '<select disabled')
            val = val.replace('width-100', 'width-100 invisible_input')
        table += "<tr><td>%s</td><td>%s</td></tr>\n" % (kname, val)
    table += "</table>"
    return table

def genid(kwds):
    "Generate id for given field"
    if  isinstance(kwds, dict):
        record = dict(kwds)
        data = json.JSONEncoder(sort_keys=True).encode(record)
    else:
        data = str(kwds)  # it is fine both in py2 and in py3
    keyhash = hashlib.md5()
    keyhash.update(encodeUnicodeToBytesConditional(data, condition=PY3))
    return keyhash.hexdigest()

def checkarg(kwds, arg):
    """Check arg in a dict that it has str/unicode type"""
    data = kwds.get(arg, None)
    cond = data and isinstance(data, (newstr, bytes))
    return cond

def checkargs(supported):
    """
    Decorator to check arguments in provided supported list
    """
    def wrap(func):
        """Wrap input function"""

        def require_string(val):
            """Check that provided input is a string"""
            if not isinstance(val, (newstr, bytes)):
                code = web_code('Invalid input')
                raise URLError('code=%s' % code)

        def wrapped_f(self, *args, **kwds):
            """Wrap function arguments"""
            # check request headers. For methods POST/PUT
            # we need to read request body to get parameters
            if  cherrypy.request.method == 'POST' or\
                cherrypy.request.method == 'PUT':
                try:
                    body = cherrypy.request.body.read()
                except:
                    body = None
                if  args and kwds:
                    code = web_code('Misleading request')
                    raise URLError('code=%s' % code)
                if  body:
                    jsondict = json.loads(body, encoding='latin-1')
                else:
                    jsondict = kwds
                for key, val in viewitems(jsondict):
                    kwds[str(key)] = str(val)

            if  not kwds:
                if  args:
                    kwds = args[-1]
            keys = []
            if  not isinstance(kwds, dict):
                code  = web_code('Unsupported kwds')
                raise URLError('code=%s' % code)
            if  kwds:
                keys = [i for i in kwds if i not in supported]
            if  keys:
                code  = web_code('Unsupported key')
                raise URLError('code=%s' % code)
            if  checkarg(kwds, 'status'):
                if  kwds['status'] not in \
                        ['new', 'assigned']:
                    code  = web_code('Unsupported view')
                    raise URLError('code=%s' % code)
            data = func (self, *args, **kwds)
            return data
        wrapped_f.__doc__  = func.__doc__
        wrapped_f.__name__ = func.__name__
        wrapped_f.exposed  = True
        return wrapped_f
    wrap.exposed = True
    return wrap

WEB_CODES = [
        (0  , 'N/A'),
        (1  , 'Unsupported key'),
        (2  , 'Unsupported value'),
        (3  , 'Unsupported method'),
        (4  , 'Unsupported collection'),
        (5  , 'Unsupported database'),
        (6  , 'Unsupported view'),
        (7  , 'Unsupported format'),
        (8  , 'Wrong type'),
        (9  , 'Misleading request'),
        (10 , 'Invalid query'),
        (11 , 'Exception'),
        (12 , 'Invalid input'),
        (13 , 'Unsupported expire value'),
        (14 , 'Unsupported order value'),
        (15 , 'Unsupported skey value'),
        (16 , 'Unsupported idx value'),
        (17 , 'Unsupported limit value'),
        (18 , 'Unsupported dir value'),
        (19 , 'Unsupported sort value'),
        (20 , 'Unsupported ajax value'),
        (21 , 'Unsupported show value'),
        (22 , 'Unsupported dasquery value'),
        (23 , 'Unsupported dbcoll value'),
        (24 , 'Unsupported msg value'),
        (25 , 'Unable to start DASCore'),
        (26 , 'No file id'),
        (27 , 'Unsupported id value'),
        (28 , 'Server error'),
        (29 , 'Query is not suitable for this view'),
        (30 , 'Parser error'),
        (31 , 'Unsupported pid value'),
        (32 , 'Unsupported interval value'),
        (33 , 'Unsupported kwds'),
]
def decode_code(code):
    """Return human readable string for provided code ID"""
    for idx, msg in WEB_CODES:
        if  code == idx:
            return msg
    return 'N/A'

def web_code(error):
    """Return WEB code for provided error string"""
    for idx, msg in WEB_CODES:
        if  msg.lower() == error.lower():
            return idx
    return -1

def sort(docs, sortby):
    "Sort given documents by sortby attribute"
    for doc in docs:
        yield doc

def reorder_list(org_list, selected):
    """
    if the first is in the list.
    move the first in front of the list
    if not, add first to the list
    """
    if isinstance(selected, list) and len(selected) == 0:
        return org_list, selected
    if not isinstance(selected, list):
        selected = [selected]
    new_list = list(org_list)
    for item in selected:
        try:
            new_list.remove(item)
        except ValueError:
            pass
    updated_list = list(selected)
    updated_list.extend(new_list)
    return updated_list, selected



#!/usr/bin/env python
"""
_ListOpenByName_

MySQL implementation of Fileset.ListOpenByName
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class ListOpenByName(DBFormatter):
    sql = "SELECT name FROM wmbs_fileset WHERE open = 1 AND name LIKE :name"

    def format(self, results):
        """
        _format_

        Take the array of rows that were returned by the query and format that
        into a single list of open fileset names.
        """
        results = DBFormatter.format(self, results)
        openFilesetNames = []

        for result in results:
            openFilesetNames.append(str(result[0]))

        return openFilesetNames

    def execute(self, name, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, binds = {"name": name},
                                      conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
The WorkflowUpdater poller component.
Among the actions performed by this component, we can list:
* find active workflows in the agent
* filter those that require pileup dataset
* find out the current location for the pileup datasets
* get a list of blocks available and locked by WM
* match those blocks with the current pileup config json file. In other words,
  blocks that are no longer locked and/or available need to be removed from the
  json file.
* update this json in the workflow sandbox
"""

import os
import json
import logging
import tarfile
import tempfile
import threading

from Utils.CertTools import cert, ckey
from Utils.IteratorTools import flattenList
from Utils.FileTools import tarMode, findFiles
from Utils.Timers import timeFunction, CodeTimer
from WMCore.Services.MSPileup.MSPileupUtils import getPileupDocs
from WMCore.Services.Rucio.Rucio import Rucio
from WMCore.WMException import WMException
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.DBS.DBSConcurrency import getBlockInfo4PU


def findJsonSandboxFiles(tfile):
    """
    Find location of sandbox JSON files
    :param tfile: sandbox tar file
    :return: list of file names
    """
    files = []
    mode = tarMode(tfile, 'r')
    with tarfile.open(tfile, mode, encoding='utf-8') as tar:
        for tarInfo in tar.getmembers():
            if tarInfo.name.endswith("pileupconf.json"):
                files.append(tarInfo.name)
    return files


def findPUName(puJsonContent):
    """
    Finds pileup name in pileup conf json file
    :param puJsonContent: content of pileupconf.json file
    :return: string (pileup name)
    """
    jsonPUName = ""
    for _puType, blocks in puJsonContent.items():
        for blockName, _content in blocks.items():
            jsonPUName = blockName.split("#")[0]
            return jsonPUName
    return jsonPUName


def extractPileupconf(tfile, fname):
    """
    Extract content of given file name from sandbox tar file
    :param tfile: sandbox tar file
    :param fname: name of the file to extract
    :return: content of file
    """
    mode = tarMode(tfile, 'r')
    with tarfile.open(tfile, mode, encoding='utf-8') as tar:
        f = tar.extractfile(fname)
        data = f.read()
        # convert our data bytes into JSON object
        return json.loads(data)


def blockLocations(jdoc):
    """
    Return dict block names and their location from provided json sandbox file.

    json structure is
    {"<type: mc>": {"<blockName>": {"FileList" [], "NumberOfEvents":1, "PhEDExNodeNames": []},
                    "<blockName>": {"FileList" [], "NumberOfEvents":1, "PhEDExNodeNames": []}, ...}

    :param jdoc: JSON document
    :return: dict {'blockName': [rses], ...}
    """
    bdict = {}
    for rec in jdoc.values():
        for key in rec.keys():
            doc = rec[key]
            bdict[key] = doc['PhEDExNodeNames']
    return bdict


def checkChanges(jdict, msPUBlockLoc):
    """
    Compare if provided json block locations are different from block locations presented
    in pileup configuration JSON file.

    :param jdict: block structure from pileupconf.json file, e.g. {'block': [rses], ...}
    :param msPUBlockLoc: block location dict obtained MSPileup service, e.g. {'block': [rses], ....}
    :return: boolean (i.e. if there are changes or not)
    """
    if sorted(jdict.keys()) != sorted(msPUBlockLoc.keys()):
        return True
    for block, msRSEs in msPUBlockLoc.items():
        if block in jdict:
            # first use-case: check for RSE match
            jsonRSEs = jdict[block]
            if sorted(msRSEs) != sorted(jsonRSEs):
                return True
        else:
            # second use-case: MSPileup has a block which is not in configuration json
            return True
    return False


def updateBlockInfo(jdoc, msPUBlockLoc, dbsUrl, logger):
    """
    Update block information within sandbox json pileup conf file using MSPileup block location info.
    The main logic is the following:
    - compare json conf blocks with those found in msPUBlockLoc
      - if block is found in msPUBlockLog we keep it and update its rses from ones found in msPUBlockLoc entry
      - if block does not exists in msPUBLockLoc we discard
    - if block from msPUBLockLoc is not found in json conf block list
      - we add this block to json conf block list and we need to have another github issue
      how to find FileList and "NumberOfEvents"

    NOTE: this function may consume lots of memory due to jdoc structure which
    we need to parse and create new dict since we can't discard keys in place
    of jdoc internal dictionary, i.e.

    jdoc structure is
    {"<type: mc>": {"<block>": {"FileList" [], "NumberOfEvents":1, "PhEDExNodeNames": []},
                    "<block>": {"FileList" [], "NumberOfEvents":1, "PhEDExNodeNames": []}, ...}
    and we need to discard block names. They are located in internal dictionary
    and in place pop-up of jdoc is not allowed since it will change size of internal dictionary.
    Therefore, to overcome this obstacles we need to copy relevant pileupName key:values
    into new dict structure and return new dict.

    :param jdoc: JSON sandbox dictionary
    :param msPUBlockLOck: dict of block with rses from MSPileup service, i.e. {'block': [rses], ... }
    :param dbsUrl: dbs Url
    :param logger: logger object
    :return: newly constructed dict
    """
    returnDict = {}
    blocksToUpdate = []
    for puType in jdoc.keys():
        newDict = {}
        # first use-case, check json blocks against MSPileup ones, if found update the rses
        for blockName in list(jdoc[puType].keys()):
            if blockName in msPUBlockLoc.keys():
                # update record rses from mspileup record
                record = jdoc[puType][blockName]
                logger.info("Block %s has locations %s, updating to %s from MSPileup",
                            blockName, record["PhEDExNodeNames"], msPUBlockLoc[blockName])
                record["PhEDExNodeNames"] = msPUBlockLoc[blockName]
                newDict[blockName] = record

        # second use-case: add MSPileup block record if they are not found in json configuration
        blocksToUpdate = []  # list of blocks to update with DBS information
        for blockName in msPUBlockLoc.keys():
            if blockName not in list(jdoc[puType].keys()):
                # add block record from MSPileup
                record = {blockName: {"PhEDExNodeNames": msPUBlockLoc[blockName]}}
                newDict[blockName] = record
                # keep track of blocks we need to update with DBS information
                blocksToUpdate.append(blockName)
        if newDict:
            returnDict[puType] = newDict

    # Update block records with DBS information.
    # Optimization note: this step is done outside of main loop above since
    # it requires fetching block information from DBS which may not be required
    # if we do not have such blocks.
    if len(blocksToUpdate) > 0:
        logger.info("Adding %s blocks from MSPileup which are not present in pileupconf.json",
                    len(blocksToUpdate))
        binfo = getBlockInfo4PU(blocksToUpdate, dbsUrl, ckey(), cert())
        for puType in returnDict.keys():
            for blockName in blocksToUpdate:
                # update block record in-place
                record = returnDict[puType][blockName]
                record.update(binfo[blockName])
    return returnDict


def writePileupJson(tfile, jdict, logger, dest=None):
    """
    Write pileup JSON sandbox files back to file system
    :param tfile: tar ball file name
    :param jdict: JSON sandbox dictionary in form: {"path-to-json1": jdoc1, "path-to-json2": jdoc2}
    :param dest: optional destination parameter to write final tar ball (use for unit tests)
    :return: nothing
    """
    bname = os.path.basename(tfile)
    dname = os.path.dirname(tfile)
    ofile = f"{dname}/new-{bname}"
    with tempfile.TemporaryDirectory() as tmpDir:
        # extract tar ball content into temporary directory
        with tarfile.open(tfile, tarMode(tfile, 'r'), encoding='utf-8') as tar:
            tar.extractall(path=tmpDir)
        # overwrite json sanbox file in temporary directory
        for jname in jdict:
            fname = os.path.join(tmpDir, jname)
            fstat = os.stat(fname)
            logger.info(f"Updating pileup file at {jname} for workflow tarball: {tfile}")
            with open(fname, 'w', encoding='utf-8') as ostream:
                json.dump(jdict[jname], ostream)
            if fstat == os.stat(fname):
                # something wrong as we did not update the file
                msg = f"File {fname} was not properly updated in {tfile}, file stat is identical"
                raise Exception(msg)
        # archive back sandbox
        with tarfile.open(ofile, tarMode(ofile, 'w'), encoding='utf-8') as tar:
            tar.add(tmpDir, arcname='')
        # overwrite existing tarball with new one
        if dest:
            os.rename(ofile, dest)
        else:
            os.rename(ofile, tfile)


class WorkflowUpdaterException(WMException):
    """
    Specific WorkflowUpdaterPoller exception handling.
    """


class WorkflowUpdaterPoller(BaseWorkerThread):
    """
    Poller that does the actual work for updating workflows.
    """

    def __init__(self, config):
        """
        Initialize WorkflowUpdaterPoller object
        :param config: a Configuration object with the component configuration
        """
        BaseWorkerThread.__init__(self)

        myThread = threading.currentThread()
        self.logger = myThread.logger
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.listActiveWflows = self.daoFactory(classname="Workflow.GetUnfinishedWorkflows")

        # parse mandatory attributes from the configuration
        self.config = config
        dbsUrl = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"
        self.dbsUrl = getattr(config.WorkflowUpdater, "dbsUrl", dbsUrl)
        if not self.dbsUrl:
            self.dbsUrl = dbsUrl
        self.rucioAcct = getattr(config.WorkflowUpdater, "rucioAccount")
        self.rucioUrl = getattr(config.WorkflowUpdater, "rucioUrl")
        self.rucioAuthUrl = getattr(config.WorkflowUpdater, "rucioAuthUrl")
        self.rucioCustomScope = getattr(config.WorkflowUpdater, "rucioCustomScope",
                                        "group.wmcore")
        self.msPileupUrl = getattr(config.WorkflowUpdater, "msPileupUrl")

        self.userCert = cert()
        self.userKey = ckey()
        self.rucio = Rucio(acct=self.rucioAcct,
                           hostUrl=self.rucioUrl,
                           authUrl=self.rucioAuthUrl)

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Executed in every polling cycle. The actual logic of the component is:
          1. find active workflows in the agent
          2. check if those active workflows are using pileup data
        :param parameters: not really used. But keeping same signature as
            the one defined in the super class.
        :return: only what is returned by the decorator
        """
        logging.info("Running Workflow updater injector poller algorithm...")
        try:
            # retrieve list of workflows with unfinished Production
            # or Processing subscriptions
            wflowSpecs = self.listActiveWflows.execute()
            if not wflowSpecs:
                logging.info("Agent has no active workflows at the moment")
                return

            # figure out workflows that have pileup
            puWflows = self.findWflowsWithPileup(wflowSpecs)
            if not puWflows:
                logging.info("Agent has no active workflows with pileup at the moment")
                return
            # resolve unique active pileup dataset names
            uniqueActivePU = set(flattenList([item['pileup'] for item in puWflows]))

            # otherwise, move on retrieving pileups
            msPileupList = self.getPileupDocs()

            # and resolve blocks in each container being used by workflows
            # considerations for 2024 are around 100 pileups each taking 2 seconds in Rucio
            with CodeTimer("Rucio block resolution", logger=logging):
                self.findRucioBlocks(uniqueActivePU, msPileupList)

            with CodeTimer("Adjust JSON spec", logger=logging):
                self.adjustJSONSpec(puWflows, msPileupList)
        except Exception as ex:
            msg = f"Caught unexpected exception in WorkflowUpdater. Details:\n{str(ex)}"
            logging.exception(msg)
            raise WorkflowUpdaterException(msg) from None

    def adjustJSONSpec(self, puWflows, msPileupList, dest=None):
        """
        Main logic of the algorithm:
        - for every pileup record find out location of tarball
        - get configuration files within tarball
        - extract block information
        - replace pileupconf.json files within tarball
        :param puWflows: list of active pileup workflows with the following structure:
            {"name": string with workflow name,
             "spec": string with spec path, e.g.
             /path/install/wmagentpy3/WorkQueueManager/cache/<workflow_name>/WMSandbox/WMWorkload.pkl
             "pileup": list of strings with pileup names}
        :param msPileupList: list of all pileup records in MSPileup service. It has the following structure:
            {"pileupName": string with pileup name,
             "customName": string with custom pileup name - if any,
             "rses": list of RSE names,
             "blocks": list of block names}
        :param dest: optional destination parameter to write final tar ball (use for unit tests)
        :return: nothing, it performs checks and adjust in place pileupconf.json file(s)
        """
        # loop over active pileup workflows
        for wflow in puWflows:
            # this logic implies that we have untarred workflow tar ball
            sandboxDir = wflow['spec'].split('/WMSandbox')[0]
            tarName = wflow['name'] + '-Sandbox.tar.bz2'
            tarFile = os.path.join(sandboxDir, tarName)
            self.logger.info("Processing workflow %s, sandbox: %s", wflow, sandboxDir)

            # find the pileup files
            puFiles = findFiles(sandboxDir, "pileupconf.json")
            self.logger.debug("### puFiles %s", puFiles)

            # json dictionary to write back to our tarball, it has the following structure
            # {pileupConfFile: pileupConfContent, ...}
            # for example
            # {'relative/path/pileupconf.json': {...}}
            jdict = {}

            # relative path name to our pileup configuration json file
            puConfJson = ""
            # list over each pileupconf.json and update it
            for puFile in puFiles:
                # load the JSON and figure out the dataset name
                with open(puFile, 'r', encoding='utf-8') as istream:
                    # puJsonContent has the following data-structure:
                    # {<dtype>: {<block_name>: {"FileList": [list_of_files]}}}
                    # for example
                    # {'mc': {'/path/block#123': {"FileList: ["/file1.root", "/file2.root"]}}}
                    puJsonContent = json.load(istream)

                    # our puFile is path to pileupconf.json, we need to construct
                    # relative path within our tarball,
                    # therefore we only take WMSandbox/.../pileupconf.json part
                    puConfJson = os.path.join('WMSandbox', puFile.split('WMSandbox/')[-1])

                jsonPUName = findPUName(puJsonContent)
                if jsonPUName == "":
                    self.logger.warning("Unable to find pileup name in %s", puFile)
                    continue
                self.logger.info("Found pileup name %s under path: %s", jsonPUName, puFile)

                # now that we know the pileup name, iterate over the MSPileup docs
                for pileupDoc in msPileupList:
                    # check if active pileup workflow is found in MSPileup one
                    pileupName = pileupDoc['pileupName']
                    if pileupName == jsonPUName:
                        # then we need to check whether there are any changes or not
                        jsonBlockLoc = blockLocations(puJsonContent)

                        # construct new data-structure:
                        # - dict of blocks and rses mapping, it will be used by checkChanges
                        msPUBlockLoc = {}
                        for blk in pileupDoc["blocks"]:
                            msPUBlockLoc[blk] = pileupDoc["rses"]

                        # are the block locations different between the JSON and MSPileup?
                        # jsonBlockLoc and msPUBlockLoc have identical data-structures
                        # {block1: [rses], block2: [rses], ...}
                        if checkChanges(jsonBlockLoc, msPUBlockLoc):
                            self.logger.info("Found differences between JSON and MSPileup content.")
                            puNewJsonContent = updateBlockInfo(puJsonContent, msPUBlockLoc, self.dbsUrl, self.logger)

                            # we should update a tarball only once for each pileup name,
                            # therefore we add new entry to jdict with our pilupe conf file
                            if puNewJsonContent:
                                # we update json file if we get new pileup content
                                jdict[puConfJson] = puNewJsonContent
                                self.logger.info("Mark %s to be updated in tarball %s with a fresh pileup content",
                                                 puConfJson, tarFile)
                            else:
                                self.logger.warning("updateBlockInfo did not return any results for %s, will skip update of pileup json content", pileupName)
                        else:
                            msg = "### There are no differences between JSON and MSPileup content "
                            msg += f"for pileup name {pileupName}. Not updating anything!"
                            self.logger.info(msg)

            # replace all pileupconf.json files at once within tarball
            # please note jdict contains of json file names and their new content
            if jdict:
                self.logger.info("Write pileup configuration file %s", tarFile)
                writePileupJson(tarFile, jdict, self.logger, dest)
            self.logger.info("Done updating spec: %s\n", wflow['spec'])

    def getPileupDocs(self):
        """
        Fetch all pileup documents from MSPileup and preprocess the data.

        Note that the 'blocks' field is for the moment just a placeholder,
        as it will be populated in a later stage,

        :return: a list of dictionaries in the following format:
          {"pileupName": string with pileup name,
           "customName": string with custom pileup name - if any,
           "rses": list of RSE names,
           "blocks": list of block names}
        """
        try:
            result = getPileupDocs(self.msPileupUrl, queryDict={}, method='GET')
        except RuntimeError as e:
            raise WorkflowUpdaterException from e

        logging.info("A total of %d pileup documents have been retrieved.", len(result))
        pileupMapList = []
        for puItem in result:
            logging.info("Pileup: %s, custom name: %s, expected at: %s, but currently available at: %s",
                         puItem['pileupName'], puItem['customName'],
                         puItem['expectedRSEs'], puItem['currentRSEs'])
            thisPU = {"pileupName": puItem['pileupName'],
                      "customName": puItem['customName'],
                      "rses": puItem['currentRSEs'],
                      "blocks": []}
            pileupMapList.append(thisPU)
        return pileupMapList

    @staticmethod
    def findWflowsWithPileup(listSpecs):
        """
        Given a list of workflow names and their respective specs, load each
        one of them and filter out those that don't require any pileup dataset.
        :param listSpecs: a list of dictionary with workflow name and spec path
        :return: a list of dictionaries with workflow name, spec path and list
            of pileup datasets being used, e.g.:
            {"name": string with workflow name,
             "spec": string with spec path,
             "pileup": list of strings with pileup names}
        """
        wflowsWithPU = []
        for wflowSpec in listSpecs:
            try:
                workloadHelper = WMWorkloadHelper()
                workloadHelper.load(wflowSpec['spec'])
                pileupSpecs = workloadHelper.listPileupDatasets()
                if pileupSpecs:
                    wflowSpec['pileup'] = set()
                    for pileupN in pileupSpecs.values():
                        wflowSpec['pileup'] =  wflowSpec['pileup'] | pileupN
                    logging.info("Workflow: %s requires pileup dataset(s): %s",
                                 wflowSpec['name'], wflowSpec['pileup'])
                    wflowsWithPU.append(wflowSpec)
                else:
                    logging.info("Workflow: %s does not require any pileup", wflowSpec['name'])
            except Exception as ex:
                msg = f"Failed to load spec file for: {wflowSpec['spec']}. Details: {str(ex)}"
                logging.error(msg)
        logging.info("There are %d pileup workflows out of %d active workflows.",
                     len(wflowsWithPU), len(listSpecs))
        return wflowsWithPU

    def findRucioBlocks(self, uniquePUList, msPileupList):
        """
        Given a list of unique pileup dataset names, list all of
        their blocks in Rucio. Note that if a pileup document contains
        a customName dataset, then we need to resolve the blocks for that
        instead.
        :param uniquePUList: a list with pileup names
        :param msPileupList: a list with dictionaries from MSPileup
        :return: update the msPileupList object in place, by populating
            the 'block' field with a list of block names
        """
        for pileupItem in msPileupList:
            if pileupItem["pileupName"] not in uniquePUList:
                # no active workflow requires this pileup
                continue

            if pileupItem["customName"]:
                logging.info("Fetching blocks for custom pileup container: %s", pileupItem["customName"])
                pileupItem["blocks"] = self.rucio.getBlocksInContainer(pileupItem["customName"],
                                                                       scope=self.rucioCustomScope)
            else:
                logging.info("Fetching blocks for pileup container: %s", pileupItem["pileupName"])
                pileupItem["blocks"] = self.rucio.getBlocksInContainer(pileupItem["pileupName"], scope='cms')



#!/usr/bin/env python
"""
_InsertType_

MySQL implementation of Subscriptions.InsertType
"""




from WMCore.Database.DBFormatter import DBFormatter

class InsertType(DBFormatter):
    sql = """INSERT INTO wmbs_sub_types (name)
               SELECT :name AS name FROM DUAL WHERE NOT EXISTS
                (SELECT name FROM wmbs_sub_types WHERE name = :name)"""

    sqlUpdate = """UPDATE wmbs_sub_types SET priority = :priority
                   WHERE name = :name"""

    def execute(self, subType, priority = None,
                conn = None, transaction = False):
        self.dbi.processData(self.sql, {"name": subType}, conn = conn,
                             transaction = transaction)
        if priority is not None:
            self.dbi.processData(self.sqlUpdate, {"name" : subType,
                                                  "priority" : priority},
                                 conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_AddRunLumi_

MySQL implementation of AddRunLumi
"""

from builtins import str, bytes

from Utils.IteratorTools import grouper
from WMCore.Database.DBFormatter import DBFormatter


class AddRunLumi(DBFormatter):
    sql = """insert dbsbuffer_file_runlumi_map (filename, run, lumi, num_events)
            select id, :run, :lumi, :num_events from dbsbuffer_file
            where lfn = :lfn"""

    def getBinds(self, filename=None, runs=None):

        binds = []

        if isinstance(filename, list):
            for entry in filename:
                binds.extend(self.getBinds(filename=entry['lfn'], runs=entry['runs']))
            return binds

        if isinstance(filename, (str, bytes)):
            lfn = filename
        elif isinstance(filename, dict):
            lfn = filename('lfn')
        else:
            raise Exception("Type of filename argument is not allowed: %s" \
                            % type(filename))

        if isinstance(runs, set):
            for run in runs:
                for lumi in run:
                    binds.append({'lfn': lfn,
                                  'run': run.run,
                                  'lumi': lumi,
                                  'num_events': run.getEventsByLumi(lumi)})
        else:
            raise Exception("Type of runs argument is not allowed: %s" \
                            % type(runs))
        return binds

    def format(self, result):
        return True

    def execute(self, file=None, runs=None, conn=None, transaction=False):
        for sliceBinds in grouper(self.getBinds(file, runs), 10000):
            result = self.dbi.processData(self.sql, sliceBinds, conn=conn,
                                          transaction=transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_InsertThreshold_

This module inserts thresholds for the given sites for Oracle
"""


from WMCore.ResourceControl.MySQL.InsertThreshold import InsertThreshold as MySQLInsertThreshold

class InsertThreshold(MySQLInsertThreshold):
    pass



#!/usr/bin/env python
"""
_AcquireFiles_

MySQL implementation of Subscription.AcquireFiles
"""

from WMCore.Database.DBFormatter import DBFormatter

class AcquireFiles(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_sub_files_acquired (subscription, fileid)
               VALUES (:subscription, :fileid)"""

    availDel = """DELETE FROM wmbs_sub_files_available
                  WHERE subscription = :subscription AND
                        fileid = :fileid"""

    def execute(self, subscription = None, file = None, conn = None,
                transaction = False):
        if isinstance(file, list):
            binds = []
            for fileid in file:
                binds.append({"subscription": subscription, "fileid": fileid})
        else:
            binds = {"subscription": subscription, "fileid": file}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.availDel, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
# encoding: utf-8
"""
Group.py

Created by Dave Evans on 2010-07-14.
Copyright (c) 2010 Fermilab. All rights reserved.
"""



import json

from WMCore.GroupUser.CouchObject import CouchObject


class Group(CouchObject):
    """
    _Group_

    Dictionary object containing attributes of a group

    """
    def __init__(self, **options):
        CouchObject.__init__(self)
        self.cdb_document_data = "group"
        self.setdefault('name', None)
        self.setdefault('administrators', [])
        self.setdefault('associated_sites', {})
        self.update(options)

    document_id = property(lambda x : "group-%s" % x['name'] )
    name = property(lambda x: x['name'])



#!/usr/bin/env python
"""Map data to locations for WorkQueue"""

from future.utils import viewitems, viewvalues, listvalues
from builtins import object
from future import standard_library
standard_library.install_aliases()

from collections import defaultdict
import logging

from urllib.parse import urlparse

from WMCore.Services.DBS.DBSReader import DBSReader
from WMCore.Services.MSPileup.MSPileupUtils import getPileupDocs
from WMCore.WorkQueue.DataStructs.ACDCBlock import ACDCBlock

# TODO: Combine with existing dls so DBSreader can do this kind of thing transparently
# TODO: Known Issue: Can't have same item in multiple dbs's at the same time.


def isGlobalDBS(dbs):
    """
    Receives a DBSReader object and finds out whether it's
    pointing to Global DBS (no matter whether it's production
    or the pre-production instance).
    """
    try:
        url = urlparse(dbs.dbsURL)
        if url.hostname.startswith('cmsweb'):
            if url.path.startswith('/dbs/prod/global') or url.path.startswith('/dbs/int/global'):
                return True
    except Exception as ex:
        logging.error("Failed to find out whether DBS is Global or not. Error: %s", str(ex))
    return False


def isDataset(inputData):
    """Check whether we're handling a block or a dataset"""
    if '#' in inputData.split('/')[-1]:
        return False
    return True


class DataLocationMapper(object):
    """Map data to locations for WorkQueue"""

    def __init__(self, logger=None, **kwargs):
        self.params = kwargs
        self.logger = logger or logging.getLogger()
        self.params.setdefault('locationFrom', 'subscription')
        self.params.setdefault('incompleteBlocks', False)
        self.params.setdefault('requireBlocksSubscribed', True)
        self.params.setdefault('rucioAccount', "wmcore_transferor")
        self.params.setdefault('rucioAccountPU', "wmcore_pileup")

        validLocationFrom = ('subscription', 'location')
        if self.params['locationFrom'] not in validLocationFrom:
            msg = "Invalid value for locationFrom: '%s'. Valid values are: %s" % (
                self.params['locationFrom'], validLocationFrom)
            raise ValueError(msg)

        self.rucio = self.params.get('rucio')
        if self.params.get('cric'):
            self.cric = self.params['cric']

        # save each DBSReader instance in the class object, such that
        # the same object is not shared amongst multiple threads
        self.dbses = {}

    def __call__(self, dataItems, rucioAcct=None):
        rucioAcct = rucioAcct or self.params['rucioAccount']
        result = {}

        dataByDbs = self.organiseByDbs(dataItems)

        for dbs, dataItems in viewitems(dataByDbs):
            # if global use Rucio, else use dbs
            if "pileup" in rucioAcct:
                output = self.locationsFromMSPileup(dataItems, dbs.dbsURL)
            elif isGlobalDBS(dbs):
                output = self.locationsFromRucio(dataItems, rucioAcct)
            else:
                output = self.locationsFromDBS(dbs, dataItems)
            result[dbs] = output

        return result

    def locationsFromRucio(self, dataItems, rucioAcct):
        """
        Get data location from Rucio. Location is mapped to the actual
        sites associated with them, so PSNs are actually returned
        :param dataItems: list of datasets/blocks names
        :param rucioAcct: string with the Rucio account name to check the rules against
        :return: dictionary key'ed by the dataset/block, with a list of PSNs as value
        """
        result = defaultdict(set)
        self.logger.info("Fetching location from Rucio for account: %s", rucioAcct)
        for dataItem in dataItems:
            try:
                dataLocations = self.rucio.getDataLockedAndAvailable(name=dataItem,
                                                                     account=rucioAcct)
                # resolve the PNNs into PSNs
                result[dataItem] = self.cric.PNNstoPSNs(dataLocations)
            except Exception as ex:
                self.logger.error('Error getting block location from Rucio for %s: %s', dataItem, str(ex))

        return result

    def locationsFromDBS(self, dbs, dataItems):
        """Get data location from dbs"""
        result = defaultdict(set)
        for dataItem in dataItems:
            try:
                if isDataset(dataItem):
                    phedexNodeNames = dbs.listDatasetLocation(dataItem)
                else:
                    phedexNodeNames = dbs.listFileBlockLocation(dataItem)
                result[dataItem].update(phedexNodeNames)
            except Exception as ex:
                self.logger.error('Error getting block location from dbs for %s: %s', dataItem, str(ex))

        # convert the sets to lists
        for name, nodes in viewitems(result):
            psns = set()
            psns.update(self.cric.PNNstoPSNs(nodes))
            result[name] = list(psns)

        return result

    def locationsFromMSPileup(self, dataItems, dbsUrl):
        """
        Get data location from MSPileup.

        :param dataItems: list, list of pileup names to query
        :param dbsUrl: str, dbs url to check which dbs server
        :return: dict, dict of pileup name keys with location set values
        """
        self.logger.info(f'Fetching locations from MSPileup for {len(dataItems)} containers')

        result = defaultdict(set)
        # TODO: Fetch multiple pileups in single request
        for dataItem in dataItems:
            try:
                queryDict = {'query': {'pileupName': dataItem},
                             'filters': ['currentRSEs', 'pileupName', 'containerFraction', 'ruleIds']}
                pileupInstance = '-testbed' if 'cmsweb-testbed' in dbsUrl else '-prod'
                msPileupUrl = f"https://cmsweb{pileupInstance}.cern.ch/ms-pileup/data/pileup"
                doc = getPileupDocs(msPileupUrl, queryDict, method='POST')[0]
                self.logger.info(f'locationsFromPileup - name: {dataItem}, currentRSEs: {doc["currentRSEs"]}, containerFraction: {doc["containerFraction"]}')
                # resolve PNNs into PSNs
                result[dataItem] = self.cric.PNNstoPSNs(doc['currentRSEs'])
            except IndexError:
                self.logger.error('Did not find any pileup document for query: %s', queryDict['query'])
            except Exception as ex:
                self.logger.error('Error getting block location from MSPileup for %s: %s', dataItem, str(ex))

        return result

    def organiseByDbs(self, dataItems):
        """Sort items by dbs instances - return dict with DBSReader as key & data items as values"""
        itemsByDbs = defaultdict(list)
        for item in dataItems:
            if ACDCBlock.checkBlockName(item['name']):
                # if it is acdc block don't update location. location should be
                # inserted when block is queued and not supposed to change
                continue

            if item['dbs_url'] not in self.dbses:
                self.dbses[item['dbs_url']] = DBSReader(item['dbs_url'])
            itemsByDbs[self.dbses[item['dbs_url']]].append(item['name'])

        return itemsByDbs


class WorkQueueDataLocationMapper(DataLocationMapper):
    """WorkQueue data location functionality"""

    def __init__(self, logger, backend, **kwargs):
        self.backend = backend
        self.logger = logger
        super(WorkQueueDataLocationMapper, self).__init__(logger, **kwargs)

    def __call__(self):
        dataItems = self.backend.getActiveData()

        dataLocations = super(WorkQueueDataLocationMapper, self).__call__(dataItems)
        self.logger.info("Found %d unique input data to update location", len(dataItems))

        # elements with multiple changed data items will fail fix this, or move to store data outside element
        modified = []
        for dataMapping in viewvalues(dataLocations):
            for data, locations in viewitems(dataMapping):
                elements = self.backend.getElementsForData(data)
                for element in elements:
                    if element.get('NoInputUpdate', False):
                        continue
                    if sorted(locations) != sorted(element['Inputs'][data]):
                        self.logger.info("%s, setting location to: %s", data, locations)
                        element['Inputs'][data] = locations
                        modified.append(element)
        self.logger.info("Updating %d elements for Input location update", len(modified))
        self.backend.saveElements(*modified)

        numParents = self.updateParentLocation()
        numPileups = self.updatePileupLocation()

        return len(modified) + numParents + numPileups

    def updateParentLocation(self):
        dataItems = self.backend.getActiveParentData()

        # fullResync incorrect with multiple dbs's - fix!!!
        dataLocations = DataLocationMapper.__call__(self, dataItems)
        self.logger.info("Found %d unique parent data to update location", len(dataItems))

        # Given that there might be multiple data items to be updated
        # handle it like a dict such that element lookup becomes easier
        modified = {}
        for dataMapping in viewvalues(dataLocations):
            for data, locations in viewitems(dataMapping):
                elements = self.backend.getElementsForParentData(data)
                for element in elements:
                    if element.get('NoInputUpdate', False):
                        continue
                    if element.id in modified:
                        element = modified[element.id]
                    for pData in element['ParentData']:
                        if pData == data:
                            if sorted(locations) != sorted(element['ParentData'][pData]):
                                self.logger.info("%s, setting location to: %s", data, locations)
                                element['ParentData'][pData] = locations
                                modified[element.id] = element
                                break
        self.logger.info("Updating %d elements for Parent location update", len(modified))
        self.backend.saveElements(*listvalues(modified))

        return len(modified)

    def updatePileupLocation(self):
        dataItems = self.backend.getActivePileupData()

        # fullResync incorrect with multiple dbs's - fix!!!
        dataLocations = DataLocationMapper.__call__(self, dataItems, self.params['rucioAccountPU'])
        self.logger.info("Found %d unique pileup data to update location", len(dataItems))

        # Given that there might be multiple data items to be updated
        # handle it like a dict such that element lookup becomes easier
        modified = {}
        for dataMapping in listvalues(dataLocations):
            for data, locations in viewitems(dataMapping):
                elements = self.backend.getElementsForPileupData(data)
                self.logger.info("Found %d elements using pileup: %s", len(elements), data)
                for element in elements:
                    if element.get('NoPileupUpdate', False):
                        continue
                    if element.id in modified:
                        element = modified[element.id]
                    for pData in element['PileupData']:
                        if pData == data:
                            if sorted(locations) != sorted(element['PileupData'][pData]):
                                self.logger.info("%s, setting location to: %s", data, locations)
                                element['PileupData'][pData] = locations
                                modified[element.id] = element
                                break
        self.logger.info("Updating %d elements for Pileup location update", len(modified))
        self.backend.saveElements(*listvalues(modified))

        return len(modified)



"""
Provide functions to collect data and upload data
"""
from __future__ import print_function

from builtins import range, object

import logging

from WMCore.Lexicon import splitCouchServiceURL

"""
JobInfoByID

Retrieve information about a job from couch and format it nicely.
"""

REQUEST_NAME_PREFIX = "test-request-"
SITE_NAME_PREFIX = "T1_test-site-"
NUM_REQUESTS = 2
NUM_SITES = 3
NUM_JOBS_EACH = 1
JOB_SLOTS = 1000
COUCH_JOB_STATUS = ['queued_first', 'queued_retry', 'cooloff', 'submitted_first',
                    'submitted_retry', 'success',
                    'failure_exception', 'failure_submit',
                    'failure_create', 'failure_cancel']
QUEUE_JOB_STATUS = ['inQueue', 'inWMBS']
BATCH_JOB_STATUS = ['submitted_pending', 'submitted_running']

class LocalCouchDBData(object):

    def __init__(self, couchURL, statSummaryDB, summaryLevel, ):
        # set the connection for local couchDB call
        print("Using LocalCouchDBData Emulator")
        self.couchURL = couchURL
        self.couchURLBase, self.dbName = splitCouchServiceURL(couchURL)
        self.summaryStatsDB = None
        self.summaryLevel = summaryLevel
        logging.info("connect couch %s:  %s" % (self.couchURLBase, self.dbName))

    def getJobSummaryByWorkflowAndSite(self):
        """
        gets the job status information by workflow

        example
        {"rows":[
            {"key":['request_name1", "queued_first", "siteA"],"value":100},
            {"key":['request_name1", "queued_first", "siteB"],"value":100},
            {"key":['request_name1", "running", "siteA"],"value":100},
            {"key":['request_name1", "success", "siteB"],"value":100}\
         ]}

         and convert to
         {'request_name1': {'queue_first'; { 'siteA': 100}}
          'request_name1': {'queue_first'; { 'siteB': 100}}
         }
        """
        doc = {}
        for i in range(NUM_REQUESTS):
            request = doc['%s%s' % (REQUEST_NAME_PREFIX, i+1)] = {}
            for status in COUCH_JOB_STATUS:
                request[status] = {}
                for j in range(NUM_SITES):
                    request[status]['%s%s' % (SITE_NAME_PREFIX, j+1)] = NUM_JOBS_EACH

        return doc

    def getEventSummaryByWorkflow(self):
        """
        _getEventSummaryByWorkflow_

        Gets the event progress by workflow.
        Returns an empty dict in the emulator
        """
        return {}

    def getJobPerformanceByTaskAndSiteFromSummaryDB(self):
        return {}

    def getSkippedFilesSummaryByWorkflow(self):
        return {}


class ReqMonDBData(object):

    def __init__(self, couchURL):
        # set the connection for local couchDB call
        self.couchURL, self.dbName = splitCouchServiceURL(couchURL)

    def uploadData(self, doc):
        """
        upload to given couchURL using cert and key authentication and authorization
        """
        return {'status':'ok'}

class WMAgentDBData(object):

    def __init__(self, summaryLevel, dbi, logger):

        # interface to WMBS/BossAir db
        print("Using %s Emulator" % self.__class__)

    def getHeartBeatWarning(self):

        agentInfo = {}
        agentInfo['status'] = 'ok'
        return agentInfo

    def getBatchJobInfo(self):
        """
        TODO: need to sync the job number (num of running job should match)
        """
        doc = {}
        for i in range(NUM_REQUESTS):
            request = doc['%s%s' % (REQUEST_NAME_PREFIX, i+1)] = {}
            for status in BATCH_JOB_STATUS:
                request[status] = {}
                for j in range(NUM_SITES):
                    request[status]['%s%s' % (SITE_NAME_PREFIX, j+1)] = NUM_JOBS_EACH

        return doc

    def getJobSlotInfo(self):

        results = []
        for i in range(NUM_SITES):
            doc = {}
            doc['cms_name'] = '%s%s' % (SITE_NAME_PREFIX, i+1)
            doc['job_slots'] = JOB_SLOTS
            results.append(doc)
        return results

    def getFinishedSubscriptionByTask(self):
        return {}


#!/usr/bin/env python
"""
_LoadFiles_

MySQL implementation of Jobs.LoadFiles
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class LoadFiles(DBFormatter):
    """
    _LoadFiles_

    Retrieve all files that are associated with the given job from the
    database.
    """
    sql = "SELECT DISTINCT fileid AS file FROM wmbs_job_assoc WHERE JOB = :jobid"

    def formatDict(self, results):
        """
        _formatDict_

        Cast the file attribute to an integer, and also handle changing the
        column name in Oracle from FILEID to FILE.
        """
        formattedResults = DBFormatter.formatDict(self, results)

        dictResults = []
        for formattedResult in formattedResults:
            dictResult = {}
            if "file" in formattedResult:
                dictResult["id"] = int(formattedResult["file"])
                dictResults.append(dictResult)

        return dictResults

    def execute(self, id, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        result = self.dbi.processData(self.sql, {"jobid": id}, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_Exists_

MySQL implementation of Subscription.Exists
"""




from WMCore.Database.DBFormatter import DBFormatter

class Exists(DBFormatter):
    sql = """SELECT id FROM wmbs_subscription
             WHERE fileset = :fileset AND workflow = :workflow"""

    def format(self, result):
        result = DBFormatter.format(self, result)

        if len(result) > 0:
            return result[0][0]
        else:
            return False

    def execute(self, workflow=None, fileset=None, type = None, conn = None,
                transaction = False):
        result = self.dbi.processData(self.sql,
                                      self.getBinds(workflow = workflow,
                                                    fileset = fileset),
                                      conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
Basic interface to HTTPS requests using ssl object managers, for python >= 2.7.9.
See usage example in:
src/python/WMCore/WMSpec/Steps/Executors/DQMUpload.py
"""
from __future__ import print_function, division

import logging
import ssl
try:
    # python2
    import urllib2
    import httplib
    HTTPSHandler = urllib2.HTTPSHandler
    HTTPSConnection = httplib.HTTPSConnection
except:
    # python3
    import urllib.request
    import http.client
    HTTPSHandler = urllib.request.HTTPSHandler
    HTTPSConnection = http.client.HTTPSConnection


class HTTPSAuthHandler(HTTPSHandler):
    """
    HTTPS authentication class to provide a ssl context with the certificates.
    """
    def __init__(self, key=None, cert=None, capath='/etc/grid-security/certificates/', level=0):
        self.logger = logging.getLogger(__name__)
        if cert:
            # then create a default ssl context manager to carry the credentials.
            # It also loads the default CA certificates
            self.ctx = ssl.create_default_context()
            self.ctx.load_cert_chain(cert, keyfile=key)
            self.ctx.load_verify_locations(None, capath)

            self.logger.info("Found %d default trusted CA certificates.", len(self.ctx.get_ca_certs()))
            ### DEBUG start ###
            #for ca in self.ctx.get_ca_certs():
            #    if 'CERN' in str(ca['subject']):
            #        print("  %s" % str(ca['subject']))
            ### DEBUG end ###
            self.logger.info("SSL context manager created with the following settings:")
            self.logger.info("  check_hostname : %s", self.ctx.check_hostname)  # default to True
            self.logger.info("  options : %s", self.ctx.options)  # default to 2197947391
            self.logger.info("  protocol : %s", self.ctx.protocol)  # default to 2 (PROTOCOL_SSLv23)
            self.logger.info("  verify_flags : %s", self.ctx.verify_flags)  # default to 0 (VERIFY_DEFAULT)
            self.logger.info("  verify_mode : %s", self.ctx.verify_mode)  # default to 2 (CERT_REQUIRED)
            HTTPSHandler.__init__(self, debuglevel=level, context=self.ctx)
        else:
            self.logger.info("Certificate not provided for HTTPSHandler")
            HTTPSHandler.__init__(self, debuglevel=level)

    def get_connection(self, host, **kwargs):
        if self.ctx:
            return HTTPSConnection(host, context=self.ctx, **kwargs)
        return HTTPSConnection(host)

    def https_open(self, req):
        """
        Overwrite the default https_open.
        """
        self.logger.debug("%s method to %s", req.get_method(), req.get_full_url())
        self.logger.debug("  with the following headers: %s", req.headers)
        return self.do_open(self.get_connection, req)



#!/usr/bin/env python
"""
_GetFailedFiles_

Oracle implementation of Subscription.GetFailedFiles
"""

from WMCore.WMBS.MySQL.Subscriptions.GetFailedFiles import GetFailedFiles \
     as GetFailedFilesMySQL

class GetFailedFiles(GetFailedFilesMySQL):
    pass



from WMCore.REST.Error import RESTError

class InvalidStateTransition(RESTError):
    "The specified object is invalid."
    http_code = 400
    app_code = 1101

    def __init__(self, requestName, currentState, newState):
        RESTError.__init__(self)
        self.message = "Invalid status transition for request: %s, from '%s' to '%s'" % (requestName,
                                                                                         currentState,
                                                                                         newState)

class InvalidSpecParameterValue(RESTError):
    "The specified object is invalid."
    http_code = 400
    app_code = 1102

    def __init__(self, message):
        RESTError.__init__(self)
        self.message = "Invalid spec parameter value: %s" % message



#!/usr/bin/env python
"""
_DBS3Buffer.SetBlockClosed_

"""

from WMComponent.DBS3Buffer.MySQL.SetBlockClosed import SetBlockClosed as MySQLSetBlockClosed

class SetBlockClosed(MySQLSetBlockClosed):
    pass



from __future__ import (division, print_function)
from builtins import object


class MockPyCondorAPI(object):
    """
    Version of Services/PyCondor intended to be used with mock or unittest.mock
    """

    def __init__(self, *args, **kwargs):
        print("Using MockPyCondorAPI")

    def getCondorJobsSummary(self):
        """
        Mock a condor query for the job summary
        """
        return []

    def getCondorJobs(self, constraint='true', attrList=None, limit=-1, opts="SummaryOnly"):
        """
        Given a job/schedd constraint, return a list of jobs attributes
        or None if the query to condor fails.
        """
        return None

    def isScheddOverloaded(self):
        """check whether job limit is reached in local schedd"""
        return False



#!/usr/bin/env python
"""
_GetBlock_

Oracle implementation of DBSBufferFiles.GetBlock
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.GetBlock import GetBlock as MySQLGetBlock

class GetBlock(MySQLGetBlock):
    """
    _GetBlock_

    Oracle implementation of DBSBufferFiles.GetBlock
    """
    pass



#!/usr/bin/env python
"""
_GetInjectedWorkflows_

Oracle implementation of Workflow.GetInjectedWorkflows

"""
__all__ = []

from WMCore.WMBS.MySQL.Workflow.GetInjectedWorkflows import GetInjectedWorkflows as MySQLGetInjectedWorkflows

class GetInjectedWorkflows(MySQLGetInjectedWorkflows):
    """
    Oracle Template

    """
